r"""Definition of :py:class:`.Observable` objects used in matching and
response matrices.

Observables are ways to monitor various lattice parameters and use them in
matching and correction procedures.

Class hierarchy
---------------

:py:class:`Observable`\ (evalfun, name, target, weight, bounds, ...)
    :py:class:`RingObservable`\ (fun, ...)

    :py:func:`GlobalOpticsObservable`\ (attrname, plane, name, ...)

    :py:class:`EmittanceObservable`\ (attrname, plane, name, ...)

    :py:class:`ElementObservable`\ (fun, ...)
        :py:class:`OrbitObservable`\ (refpts, axis, name, ...)

        :py:class:`TrajectoryObservable`\ (refpts, axis, name, ...)

        :py:class:`MatrixObservable`\ (refpts, axis, name, ...)

        :py:class:`LocalOpticsObservable`\ (refpts, attrname, axis, name, ...)

        :py:class:`LatticeObservable`\ (refpts, attrname, index, name, ...)

        :py:class:`GeometryObservable`\ (refpts, attrname, name, ...)

        :py:class:`.RDTObservable`\ (refpts, RDTname, name, ...)

:py:class:`.Observable`\ s are usually not evaluated directly, but through a
container which performs the required optics computation and feeds each
:py:class:`.Observable` with its specific data. After evaluation, each
:py:class:`.Observable` provides the following properties:

- :py:attr:`~Observable.value`
- :py:attr:`~Observable.weight`
- :py:attr:`~Observable.weighted_value`: :pycode:`value / weight`
- :py:attr:`~Observable.deviation`:  :pycode:`value - target`
- :py:attr:`~Observable.residual`:  :pycode:`((value - target)/weight)**2`
"""

from __future__ import annotations

__all__ = [
    "Observable",
    "RingObservable",
    "EmittanceObservable",
    "GlobalOpticsObservable",
    "ElementObservable",
    "OrbitObservable",
    "GeometryObservable",
    "LocalOpticsObservable",
    "LatticeObservable",
    "MatrixObservable",
    "TrajectoryObservable",
    "Need",
]

from collections.abc import Callable, Set
from functools import partial
from enum import Enum
from itertools import repeat
from typing import Union

# For sys.version_info.minor < 9:
from typing import Tuple

import numpy as np
import numpy.typing as npt

from ..lattice import AtError, AxisDef, axis_, plane_
from ..lattice import Lattice, Refpts, End

RefIndex = Union[int, Tuple[int, ...], slice]


# Observables must be pickleable. For this, the evaluation function must be a
# module-level function. No inner, nested function is allowed. So nested
# functions are replaced be module-level callable class instances:


class _Convolve:

    def __init__(self, modfun, fun, *args, **kwargs):
        self.modfun = modfun
        self.fun = fun
        self.args = args
        self.kwargs = kwargs

    def __call__(self, *a):
        return self.modfun(self.fun(*a), *self.args, **self.kwargs)


class _ArrayAccess:
    """Access a selected item in an array."""

    def __init__(self, index):
        self.index = _all_rows(index)

    def __call__(self, data):
        index = self.index
        return data if index is None else data[self.index]


def _record_access(param, index, data):
    """Access a selected item in a record array"""
    val = getattr(data, param)
    return val if index is None else val[index]


def _fun_access(fun, index, data):
    """Access a selected item in the output of a user-defined function"""
    val = fun(data)
    return val if index is None else val[index]


def _muf_access(_, index, data):
    mu = _record_access("mu", index, data)
    return np.remainder(mu, 2.0 * np.pi)


def _mu2pi_access(_, index, data):
    mu = _record_access("mu", index, data)
    return mu / 2.0 / np.pi


def _mu2pif_access(_, index, data):
    mu = _record_access("mu", index, data)
    return np.remainder(mu / 2.0 / np.pi, 1.0)


_opdata = {
    "muf": _muf_access,
    "mu2pi": _mu2pi_access,
    "mu2pif": _mu2pif_access,
}

_arrayproc = {
    None: None,
    "real": np.real,
    "imag": np.imag,
    "abs": np.absolute,
    "angle": np.angle,
    "log": np.log,
    "exp": np.exp,
    "sqrt": np.sqrt,
}

_statproc = {
    None: None,
    "mean": np.mean,
    "std": np.std,
    "var": np.var,
    "min": np.min,
    "max": np.max,
}


def _all_rows(index: RefIndex | None):
    """Prepends "all rows" (":") to an index tuple."""
    if index is None:
        return None
    if isinstance(index, tuple):
        return (slice(None),) + index
    else:
        return slice(None), index


def _get_fun(fname, fdict) -> Callable:
    """Get a processing from its name"""
    if callable(fname):
        return fname
    else:
        return fdict[fname]


class _Tune:
    """Get integer tune from the phase advance."""

    def __init__(self, idx: RefIndex):
        self.fun = partial(_record_access, "mu", _all_rows(idx))

    def __call__(self, data):
        mu = self.fun(data)
        return np.squeeze(mu, axis=0) / 2.0 / np.pi


class _Ring:
    """Get an attribute of a lattice element."""

    def __init__(self, attrname, index, refpts):
        self.get_val = partial(_record_access, attrname, index)
        self.refpts = refpts

    def __call__(self, ring):
        vals = [self.get_val(el) for el in ring.select(self.refpts)]
        return np.array(vals)


class Need(Enum):
    """Defines the computation requirements for an :py:class:`Observable`."""

    #:  Provides the *ring* data to the evaluation function
    RING = 1
    #:  Specify :py:func:`.find_orbit` computation and provide its *orbit* output
    #:  to the evaluation function
    ORBIT = 2
    #:  Specify :py:func:`.find_m44` or :py:func:`.find_m44` computation and
    #:  provide its *m44* or *m66* output to the evaluation function
    MATRIX = 3
    #:  Specify :py:func:`.get_optics` computation and provide its *ringdata* output
    #:  to the evaluation function
    GLOBALOPTICS = 4
    #:  Specify :py:func:`.get_optics` computation and provide its *elemdata* output
    #:  to the evaluation function
    LOCALOPTICS = 5
    #:  Specify :py:func:`.lattice_pass` computation and provide its *r_out*
    #:  to the evaluation function
    TRAJECTORY = 6
    #:  Specify :py:func:`.envelope_parameters` computation and provide its
    #:  *params* output to the evaluation function
    EMITTANCE = 7
    #:  Associated with LOCALOPTICS, require local optics computation at all
    #:  points: slower but avoids jumps in phase advance
    ALL_POINTS = 8
    #:  Associated with LOCALOPTICS, require the *get_chrom* keyword
    CHROMATICITY = 9
    #:  Associated with LOCALOPTICS, require the *get_w* keyword
    W_FUNCTIONS = 10
    #:  Specify :py:meth:`~.Lattice.get_geometry` computation and provide its output
    #:  to the evaluation function
    GEOMETRY = 11
    #:  Specify :py:func:`RDT <.get_rdts>` computation and provide its output to
    #:  the evaluation function
    RDT = 12
    #:  Associated with RDT: request 2nd order calculation
    RDT_2ND_ORDER = 13


class Observable:
    """Base class for Observables. Can be used for user-defined observables."""

    def __init__(
        self,
        fun: Callable,
        *args,
        name: str | None = None,
        target: npt.ArrayLike | None = None,
        weight: npt.ArrayLike = 1.0,
        bounds=(0.0, 0.0),
        needs: Set[Need] | None = None,
        postfun: Callable | str | None = None,
        **kwargs,
    ):
        r"""Args:
            name:           Observable name. If :py:obj:`None`, an explicit
              name will be generated
            fun:            :ref:`evaluation function <base_eval>`
            *args:          Arguments provided to the evaluation function
            target:         Target value for a constraint. If :py:obj:`None`
              (default), the residual will always be zero.
            weight:         Weight factor: the residual is
              :pycode:`((value-target)/weight)**2`
            bounds:         Tuple of lower and upper bounds. The parameter
              is constrained in the interval
              [*target*\ +\ *low_bound* *target*\ +\ *up_bound*]
            postfun:        Post-processing function. It can be any numpy ufunc or a
              function name in {"real", "imag", "abs", "angle", "log", "exp", "sqrt"}.
            needs:          Set of requirements. This selects the data provided
              to the evaluation function. *needs* items are members of the
              :py:class:`Need` enumeration

        Keyword Args:
            **kwargs:       Keyword arguments provided to the evaluation function

        The *target*, *weight* and *bounds* inputs must be broadcastable to the
        shape of *value*.

        .. _base_eval:
        .. rubric:: User-defined evaluation function

        The general form is:

        :pycode:`value = fun(*data, *args, **kwargs)`

        - *data* depends on the *needs* argument, and by default is empty. If several
          needed values are specified, their order is: *ring*, *orbit*, *m44/m66*,
          *ringdata*, *elemdata*, *r_out*, *params*, *geomdata*,
        - *args* are the positional arguments provided to the observable constructor,
        - *kwargs* are the keyword arguments provided to the observable constructor,
        - *value* is the value of the observable.

        For user-defined evaluation functions using linear optics data or
        emittance data, it is recommended to use
        :py:class:`LocalOpticsObservable`, :py:obj:`GlobalOpticsObservable`
        or :py:class:`EmittanceObservable` which provide the corresponding
        *data* argument.
        """
        name = fun.__name__ if name is None else name
        postfun = _get_fun(postfun, _arrayproc)
        if postfun:
            name = f"{postfun.__name__}({name})"
            fun = _Convolve(postfun, fun)
        self.fun: Callable = fun  #: Evaluation function
        self.needs: Set[Need] = needs or set()  #: Set of requirements
        self.name: str = name  #: Observable name
        self.target: npt.ArrayLike | None = target  #: Target value
        self.w: npt.NDArray[float] = np.asarray(weight, dtype=float)
        self.lbound, self.ubound = bounds
        self.initial: npt.NDArray[float] | None = None
        self._value: npt.NDArray[float] | Exception | None = None
        self._shape: tuple[int, ...] | None = None
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        """Return the string representation of the Observable."""
        return "\n".join((self._header(), self._all_lines()))

    @staticmethod
    def _header():
        """Header line."""
        fstring = "\n    {:<12} {:>16}  {:>16}  {:>16}  {:>16}  {:>16} "
        return fstring.format(
            "location", "Initial", "Actual", "Low bound", "High bound", "deviation"
        )

    @staticmethod
    def _line(loc, *items):
        def pitem(v):
            if v is None:
                return f" {'-':>14}   "
            elif isinstance(v, Exception):
                return f" {type(v).__name__:>16} "
            elif isinstance(v, np.ndarray) and v.ndim > 0:
                if v.size == 0:
                    return f" {'[]':16} "
                else:
                    return f" [{v.flat[0]:< 10.4} ...] "
            else:
                return f" {v: 16.6} "

        its = [pitem(v) for v in items]
        return f"    {loc:<12}" + "".join(its)

    def _all_lines(self):
        vnow = self._value
        if vnow is None or isinstance(vnow, Exception):
            deviation = None
            vmin = None
            vmax = None
        else:
            deviation = self.deviation
            if self.target is None:
                vmin = None
                vmax = None
            else:
                target = np.broadcast_to(self.target, vnow.shape)  # type: ignore
                vmin = target + self.lbound
                vmax = target + self.ubound
        values = self._line("", self.initial, vnow, vmin, vmax, deviation)
        return "\n".join((self.name, values))

    def _setup(self, ring: Lattice):
        """Setup function called when the observable is added to a list."""
        pass

    def evaluate(self, *data, initial: bool = False) -> npt.NDArray[float] | Exception:
        """Compute and store the value of the observable.

        The direct evaluation of a single :py:class:`Observable` is normally
        not used. This method is called by the :py:class:`.ObservableList`
        container which provides the *data* arguments.

        Args:
            *data:      Raw data, provided by :py:class:`.ObservableList` and
              sent to the evaluation function
            initial:    It :py:obj:`None`, store the result as the initial
              value

        Returns:
            value:      The value of the observable or the error in evaluation
        """
        for d in data:
            if isinstance(d, Exception):
                message = f"Evaluation of {self.name} failed: {d.args[0]}"
                err = type(d)(message).with_traceback(d.__traceback__)
                self._value = err
                return err

        val = np.asarray(self.fun(*data, *self.args, **self.kwargs))
        if initial:
            self.initial = val
        self._shape = val.shape
        self._value = val
        return val

    def check(self) -> bool:
        """Check if evaluation is done

        Returns:
            ok: :py:obj:`True` if evaluation is done, :py:obj:`False` otherwise.

        Raises:
            AtError:    if the value is doubtful: evaluation failed, empty value…
        """
        return self.value is not None

    @staticmethod
    def check_value(value: npt.NDArray[float] | Exception) -> npt.NDArray[float]:
        if isinstance(value, Exception):
            raise type(value)(value.args[0]) from value
        return value

    @property
    def value(self) -> npt.NDArray[float]:
        """Value of the observable."""
        return self.check_value(self._value)

    @property
    def weight(self) -> npt.NDArray[float]:
        """Observable weight."""
        return np.broadcast_to(self.w, self._value.shape)  # type: ignore

    @weight.setter
    def weight(self, w: npt.ArrayLike):
        self.w = np.asarray(w, dtype=float)

    @property
    def weighted_value(self) -> npt.NDArray[float]:
        """Weighted value of the Observable, computed as
        :pycode:`weighted_value = value/weight`.
        """
        return self.value / self.w

    @property
    def deviation(self) -> npt.NDArray[float]:
        """Deviation from target value, computed as
        :pycode:`deviation = value-target`.

        If *target* is :py:obj:`None`, the deviation is zero for any value.
        """
        vnow = self.value
        if vnow is None:
            deviation = None
        elif self.target is None:
            deviation = np.broadcast_to(0.0, vnow.shape)
        else:
            vsh = vnow.shape
            diff = np.atleast_1d(vnow - np.broadcast_to(self.target, vsh))
            lb = diff - self.lbound
            ub = diff - self.ubound
            lb[lb >= 0] = 0
            ub[ub <= 0] = 0
            deviation = (lb + ub).reshape(vsh)
        return deviation

    @property
    def weighted_deviation(self) -> npt.NDArray[float]:
        """:pycode:`weighted_deviation = (value-target)/weight`.

        If *target* is :py:obj:`None`, the weighted deviation is zero for any value.
        """
        return self.deviation / self.w

    @property
    def residual(self) -> npt.NDArray[float]:
        """residual, computed as :pycode:`residual = ((value-target)/weight)**2`.

        If *target* is :py:obj:`None`, the residual is zero for any value.
        """
        # absolute necessary for complex data
        return np.absolute(self.weighted_deviation) ** 2

    @staticmethod
    def _set_name(name, param, index):
        """Compute a default observable names."""
        if name is None:
            if (
                index is Ellipsis
                or index is None
                or (isinstance(index, str) and index in {":", "..."})
            ):
                subscript = ""
            elif isinstance(index, tuple):
                ids = ", ".join(str(k) for k in index)
                subscript = f"[{ids}]"
            else:
                subscript = f"[{index}]"
            if callable(param):
                try:
                    base = param.__name__
                except AttributeError:
                    base = "<function>"
            else:
                base = param
            name = base + subscript
        return name


class RingObservable(Observable):
    """Observe any user-defined property of a ring."""

    def __init__(
        self,
        fun: Callable,
        name: str | None = None,
        **kwargs,
    ):
        r"""Args:
            fun:            :ref:`user-defined evaluation function <ring_eval>`
            name:           Observable name. If :py:obj:`None`, an explicit
              name will be generated.

        Keyword Args:
            target:         Target value for a constraint. If :py:obj:`None`
              (default), the residual will always be zero.
            weight:         Weight factor: the residual is
              :pycode:`((value-target)/weight)**2`
            bounds:         Tuple of lower and upper bounds. The parameter
              is constrained in the interval
              [*target*\ +\ *low_bound* *target*\ +\ *up_bound*]
            postfun:        Post-processing function. It can be any numpy ufunc or a
              function name in {"real", "imag", "abs", "angle", "log", "exp", "sqrt"}.

        The *target*, *weight* and *bounds* inputs must be broadcastable to the
        shape of *value*.

        .. _ring_eval:
        .. rubric:: User-defined evaluation function

        It is called as:

        :pycode:`value = fun(ring)`

        - *ring* is the lattice description,
        - *value* is the value of the Observable.

        Examples:

            >>> def circumference(ring):
            ...     return ring.get_s_pos(len(ring))[0]
            >>> obs = RingObservable(circumference)

            Defines an Observable for the ring circumference.

            >>> def momentum_compaction(ring):
            ...     return ring.get_mcf()
            >>> obs = RingObservable(momentum_compaction)

            Defines an Observable for the momentum compaction factor.
        """
        needs = {Need.RING}
        name = self._set_name(name, fun, None)
        super().__init__(fun, name=name, needs=needs, **kwargs)


class ElementObservable(Observable):
    """Base class for Observables linked to a position in the lattice."""

    def __init__(
        self,
        fun: Callable,
        refpts: Refpts,
        name: str | None = None,
        statfun: Callable | str | None = None,
        postfun: Callable | str | None = None,
        **kwargs,
    ):
        r"""Args:
            fun:            :ref:`evaluation function <base_eval>`
            refpts:         Observation points.
              See ":ref:`Selecting elements in a lattice <refpts>`"
            name:           Observable name. If :py:obj:`None`, an explicit
              name will be generated
            postfun:        Post-processing function. It can be any numpy ufunc or a
              function name in {"real", "imag", "abs", "angle", "log", "exp", "sqrt"}.
            statfun:        Statistics post-processing function. it can be a numpy
              function or a function name in {"mean", "std", "var", "min", "max"}.
              Example: :pycode:`statfun=numpy.mean`.

        Keyword Args:
            target:         Target value for a constraint. If :py:obj:`None`
              (default), the residual will always be zero.
            weight:         Weight factor: the residual is
              :pycode:`((value-target)/weight)**2`
            bounds:         Tuple of lower and upper bounds. The parameter
              is constrained in the interval
              [*target*\ +\ *low_bound* *target*\ +\ *up_bound*]

        .. rubric:: Shape of the value

        A :pycode:`nrefs` dimension, with :pycode:`nrefs` the number
        of reference points, is prepended to the shape of the output of *fun*,
        **even if nrefs == 1**. The *target*, *weight* and *bounds* inputs must be
        broadcastable to the shape of *value*.
        """
        name = fun.__name__ if name is None else name
        postfun = _get_fun(postfun, _arrayproc)
        if postfun:
            name = f"{postfun.__name__}({name})"
            fun = _Convolve(postfun, fun)
        statfun = _get_fun(statfun, _statproc)
        if statfun:
            summary = kwargs.pop("summary", True)
            name = f"{statfun.__name__}({name})"
            fun = _Convolve(statfun, fun, axis=0)
        else:
            summary = kwargs.pop("summary", False)
        super().__init__(fun, name=name, **kwargs)
        self.summary = summary
        self.refpts = refpts
        self._boolrefs = None
        self._excluded = None
        self._locations = [""]

    def check(self) -> bool:
        ok = super().check()
        shp = self._shape
        if ok and shp and shp[0] <= 0:
            raise AtError(
                f"Observable {self.name!r}: No location selected in the lattice."
            )
        return ok

    def _all_lines(self):
        if self.summary:
            return super()._all_lines()
        else:
            vnow = self._value
            if vnow is None or isinstance(vnow, Exception):
                vnow = repeat(vnow)
                deviation = repeat(None)
                vmin = repeat(None)
                vmax = repeat(None)
            else:
                deviation = self.deviation
                if self.target is None:
                    vmin = repeat(None)
                    vmax = repeat(None)
                else:
                    target = np.broadcast_to(self.target, vnow.shape)  # type: ignore
                    vmin = target + self.lbound
                    vmax = target + self.ubound
            vini = self.initial
            if vini is None:
                vini = repeat(None)
            viter = zip(self._locations, vini, vnow, vmin, vmax, deviation)
            values = "\n".join(self._line(*vv) for vv in viter)
            return "\n".join((self.name, values))

    def _setup(self, ring: Lattice):
        boolrefs = ring.get_bool_index(self.refpts)
        excluded = ring.get_bool_index(self._excluded)
        boolrefs &= ~excluded
        self._boolrefs = boolrefs
        locs = [el.FamName for el in ring.select(self._boolrefs[:-1])]
        if boolrefs[-1]:
            locs.append("End")
        self._locations = locs


class GeometryObservable(ElementObservable):
    """Observe the geometrical parameters of the reference trajectory.

    Process the *geomdata* output of :py:func:`.get_geometry`.
    """

    _field_list = {"x", "y", "angle"}

    def __init__(self, refpts: Refpts, param: str, name: str | None = None, **kwargs):
        # noinspection PyUnresolvedReferences
        r"""Args:
            refpts:         Observation points.
              See ":ref:`Selecting elements in a lattice <refpts>`"
            param:          Geometry parameter name: one in {'x', 'y', 'angle'}
            name:           Observable name. If :py:obj:`None`, an explicit
              name will be generated.

        Keyword Args:
            statfun:        Post-processing function called on the value of the
              observable. Example: :pycode:`statfun=numpy.mean`
            target:         Target value for a constraint. If :py:obj:`None`
              (default), the residual will always be zero.
            weight:         Weight factor: the residual is
              :pycode:`((value-target)/weight)**2`
            bounds:         Tuple of lower and upper bounds. The parameter
              is constrained in the interval
              [*target*\ +\ *low_bound* *target*\ +\ *up_bound*]

        The *target*, *weight* and *bounds* inputs must be broadcastable to the
        shape of *value*.

        Example:

            >>> obs = GeometryObservable(at.Monitor, param="x")

            Observe x coordinate of monitors
        """
        if param not in self._field_list:
            raise ValueError(f"Expected {param!r} to be one of {self._field_list!r}")
        name = self._set_name(name, "geometry", param)
        fun = partial(_record_access, param, None)
        needs = {Need.GEOMETRY}
        super().__init__(fun, refpts, needs=needs, name=name, **kwargs)


class OrbitObservable(ElementObservable):
    """Observe the closed orbit coordinates at selected locations.

    Process the *orbit* output of :py:func:`.find_orbit`.
    """

    def __init__(
        self, refpts: Refpts, axis: AxisDef = None, name: str | None = None, **kwargs
    ):
        # noinspection PyUnresolvedReferences
        r"""Args:
            refpts:         Observation points.
              See ":ref:`Selecting elements in a lattice <refpts>`"
            axis:           Index in the orbit vector, If :py:obj:`None`,
              the whole vector is specified
            name:           Observable name. If :py:obj:`None`, an explicit
              name will be generated.

        Keyword Args:
            postfun:        Post-processing function. It can be any numpy ufunc or a
              function name in {"real", "imag", "abs", "angle", "log", "exp", "sqrt"}.
            statfun:        Statistics post-processing function. it can be a numpy
              function or a function name in {"mean", "std", "var", "min", "max"}.
              Example: :pycode:`statfun=numpy.mean`.
            target:         Target value for a constraint. If :py:obj:`None`
              (default), the residual will always be zero.
            weight:         Weight factor: the residual is
              :pycode:`((value-target)/weight)**2`
            bounds:         Tuple of lower and upper bounds. The parameter
              is constrained in the interval
              [*target*\ +\ *low_bound* *target*\ +\ *up_bound*]

        .. rubric:: Shape of the value

        If *axis* is :py:obj:`None` (whole orbit vector), then *value* has shape
        :pycode:`(nrefs, 6)` with :pycode:`nrefs` number of reference points: a
        :pycode:`nrefs` dimension is prepended to the shape of the orbit vector,
        **even if nrefs == 1**. A single coordinate has shape :pycode:`(nrefs,)`.
        The *target*, *weight* and *bounds* inputs must be broadcastable to the shape
        of *value*.

        Example:
            >>> obs = OrbitObservable(at.Monitor, axis=0)

            Observe the horizontal closed orbit at monitor locations
        """
        name = self._set_name(name, "orbit", axis_(axis, key="code"))
        fun = _ArrayAccess(axis_(axis, key="index"))
        needs = {Need.ORBIT}
        super().__init__(fun, refpts, needs=needs, name=name, **kwargs)


class MatrixObservable(ElementObservable):
    """Observe the closed orbit at selected locations.

    Processs the result of calling :py:func:`.find_m44` or :py:func:`.find_m44`
    depending upon :py:meth:`~.Lattice.is_6d`.
    """

    def __init__(
        self,
        refpts: Refpts,
        axis: AxisDef = Ellipsis,
        name: str | None = None,
        **kwargs,
    ):
        # noinspection PyUnresolvedReferences
        r"""Args:
            refpts:         Observation points.
              See ":ref:`Selecting elements in a lattice <refpts>`"
            axis:           Index in the transfer matrix, If :py:obj:`Ellipsis`,
              the whole matrix is specified
            name:           Observable name. If :py:obj:`None`, an explicit
              name will be generated.

        Keyword Args:
            postfun:        Post-processing function. It can be any numpy ufunc or a
              function name in {"real", "imag", "abs", "angle", "log", "exp", "sqrt"}.
            statfun:        Statistics post-processing function. it can be a numpy
              function or a function name in {"mean", "std", "var", "min", "max"}.
              Example: :pycode:`statfun=numpy.mean`.
            target:         Target value for a constraint. If :py:obj:`None`
              (default), the residual will always be zero.
            weight:         Weight factor: the residual is
              :pycode:`((value-target)/weight)**2`
            bounds:         Tuple of lower and upper bounds. The parameter
              is constrained in the interval
              [*target*\ +\ *low_bound* *target*\ +\ *up_bound*]

        The *target*, *weight* and *bounds* inputs must be broadcastable to the
        shape of *value*.

        Example:

            >>> obs = MatrixObservable(at.Monitor, axis=("x", "px"))

            Observe the transfer matrix from origin to monitor locations and
            extract T[0,1]
        """
        name = self._set_name(name, "matrix", axis_(axis, key="code"))
        fun = _ArrayAccess(axis_(axis, key="index"))
        needs = {Need.MATRIX}
        super().__init__(fun, refpts, needs=needs, name=name, **kwargs)


class _GlobalOpticsObservable(Observable):
    def __init__(
        self, param: str, plane: AxisDef = None, name: str | None = None, **kwargs
    ):
        # noinspection PyUnresolvedReferences
        r"""Args:
            param:          Optics parameter name (see :py:func:`.get_optics`)
              or user-defined evaluation function called as:
              :pycode:`value = fun(ringdata, ring=ring)` and returning the value of
              the Observable
            plane:          Index in the parameter array, If :py:obj:`None`,
              the whole array is specified
            name:           Observable name. If :py:obj:`None`, an explicit
              name will be generated.

        Keyword Args:
            target:         Target value for a constraint. If :py:obj:`None`
              (default), the residual will always be zero.
            weight:         Weight factor: the residual is
              :pycode:`((value-target)/weight)**2`
            bounds:         Tuple of lower and upper bounds. The parameter
              is constrained in the interval
              [*target*\ +\ *low_bound* *target*\ +\ *up_bound*]
            postfun:        Post-processing function. It can be any numpy ufunc or a
              function name in {"real", "imag", "abs", "angle", "log", "exp", "sqrt"}.

        The *target*, *weight* and *bounds* inputs must be broadcastable to the
        shape of *value*.
        """
        needs = {Need.GLOBALOPTICS}
        name = self._set_name(name, param, plane_(plane, key="code"))
        if callable(param):
            fun = partial(_fun_access, param, plane_(plane, key="index"))
            needs.add(Need.CHROMATICITY)
        else:
            fun = partial(_record_access, param, plane_(plane, key="index"))
            if param == "chromaticity":
                needs.add(Need.CHROMATICITY)
        super().__init__(fun, needs=needs, name=name, **kwargs)


class LocalOpticsObservable(ElementObservable):
    """Observe a local optics parameter at selected locations.

    Process the local output of :py:func:`.get_optics`.
    """

    def __init__(
        self,
        refpts: Refpts,
        param: str | Callable,
        plane: AxisDef = Ellipsis,
        name: str | None = None,
        all_points: bool = False,
        **kwargs,
    ):
        # noinspection PyUnresolvedReferences
        r"""Args:
            refpts:         Observation points.
              See ":ref:`Selecting elements in a lattice <refpts>`"
            param:          :ref:`Optics parameter name <localoptics_param>`
              or :ref:`user-defined evaluation function <localoptics_eval>`
            plane:          Index in the parameter array, If :py:obj:`Ellipsis`,
              the whole array is specified
            name:           Observable name. If :py:obj:`None`, an explicit
              name will be generated
            all_points:     Compute the local optics at all elements. This avoids
              discontinuities in phase advances. This is automatically set for the
              'mu' parameter, but may need to be specified for user-defined evaluation
              functions using the phase advance.

        Keyword Args:
            summary:        Set to :py:obj:`True` if the user-defined
             evaluation function returns a single item (see below)
            postfun:        Post-processing function. It can be any numpy ufunc or a
              function name in {"real", "imag", "abs", "angle", "log", "exp", "sqrt"}.
            statfun:        Statistics post-processing function. it can be a numpy
              function or a function name in {"mean", "std", "var", "min", "max"}.
              Example: :pycode:`statfun=numpy.mean`.
            target:         Target value for a constraint. If :py:obj:`None`
              (default), the residual will always be zero.
            weight:         Weight factor: the residual is
              :pycode:`((value-target)/weight)**2`
            bounds:         Tuple of lower and upper bounds. The parameter
              is constrained in the interval
              [*target*\ +\ *low_bound* *target*\ +\ *up_bound*]

        .. rubric:: Shape of the value

        If the requested attribute has shape :pycode:`shp`, then
        *value* has shape :pycode:`(nrefs,) + shp` with :pycode:`nrefs` number of
        reference points: a  :pycode:`nrefs` dimension is prepended to the shape of
        the attribute, **even if nrefs == 1**. The *target*, *weight* and *bounds*
        inputs must be broadcastable to the shape of *value*. For instance, a *target*
        with shape :pycode:`shp` will automatically broadcast and apply  to all
        reference points.

        .. _localoptics_param:
        .. rubric:: Optics parameter name

        In addition to :py:func:`.get_optics` parameter names, LocalOpticsObservable
        adds 3 parameters: *muf*, *mu2pi* and *mu2pif*:

        ================    ===================================================
        **s_pos**           longitudinal position [m]
        **M**               (6, 6) transfer matrix M from the beginning of ring
                            to the entrance of the element
        **closed_orbit**    (6,) closed orbit vector
        **dispersion**      (4,) dispersion vector
        **A**               (6, 6) A-matrix
        **R**               (3, 6, 6) R-matrices
        **beta**            :math:`\left[ \beta_x,\beta_y \right]` vector
        **alpha**           :math:`\left[ \alpha_x,\alpha_y \right]` vector
        **mu**              :math:`\left[ \mu_x,\mu_y \right]`, betatron phase
        **mu2pi**           :math:`\left[ \mu_x,\mu_y \right]/2\pi`, reduced betatron
                            phase
        **muf**             :math:`\left[ \mu_x,\mu_y \right]`, betatron phase
                            (modulo :math:`2\pi`)
        **mu2pif**          :math:`\mathrm{frac}(\left[ \mu_x,\mu_y \right]/2\pi)`,
                            fractional part of the reduced betatron phase
        **W**               :math:`\left[ W_x,W_y \right]` only if *get_w*
                            is :py:obj:`True`: chromatic amplitude function
        **Wp**              :math:`\left[ Wp_x,Wp_y \right]` only if *get_w*
                            is :py:obj:`True`: chromatic phase function
        **dalpha**          (2,) alpha derivative vector
                            (:math:`\Delta \alpha/ \delta_p`)
        **dbeta**           (2,) beta derivative vector
                            (:math:`\Delta \beta/ \delta_p`)
        **dmu**             (2,) mu derivative vector
                            (:math:`\Delta \mu/ \delta_p`)
        **ddispersion**     (4,) dispersion derivative vector
                            (:math:`\Delta D/ \delta_p`)
        **dR**              (3, 6, 6) R derivative vector
                            (:math:`\Delta R/ \delta_p`)
        ================    ===================================================

        .. _localoptics_eval:
        .. rubric:: User-defined evaluation function

        The observable value is computed as:

        :pycode:`value = fun(elemdata)[plane]`

        - *elemdata* is the output of :py:func:`.get_optics`, evaluated at the *refpts*
          of the observable,
        - *value* is the value of the Observable and must have one line per
          refpoint. Alternatively, it may be a single line, but then the
          *summary* keyword must be set to :py:obj:`True`.
        - the *plane* keyword then selects the desired values in the function output.

        Examples:

            >>> obs = LocalOpticsObservable(at.Monitor, "beta")

            Observe the beta in both planes at all :py:class:`.Monitor`
            locations

            >>> obs = LocalOpticsObservable(
            ...     at.Quadrupole, "beta", plane="y", statfun=np.max
            ... )

            Observe the maximum vertical beta in Quadrupoles

            >>> def phase_advance(elemdata):
            ...     mu = elemdata.mu
            ...     return mu[-1] - mu[0]
            >>>
            >>> allobs.append(
            ...     LocalOpticsObservable(
            ...         [33, 101], phase_advance, plane="y", all_points=True, summary=True
            ...     )
            ... )

            The user-defined evaluation function computes the phase-advance
            between the 1st and last given reference points, here the elements
            33 and 101 of the lattice
        """
        if param in {"M", "closed_orbit", "dispersion", "A", "R"}:
            ax_ = axis_
        else:
            ax_ = plane_

        needs = {Need.LOCALOPTICS}
        name = self._set_name(name, param, ax_(plane, key="code"))
        index = _all_rows(ax_(plane, key="index"))
        if callable(param):
            fun = partial(_fun_access, param, ax_(plane, key="index"))
        else:
            fun = partial(_opdata.get(param, _record_access), param, index)
            if param in {"mu", "mu2pi"} or all_points:
                needs.add(Need.ALL_POINTS)
            if param in {"W", "Wp", "dalpha", "dbeta", "dmu", "ddispersion", "dR"}:
                needs.add(Need.W_FUNCTIONS)

        super().__init__(fun, refpts, needs=needs, name=name, **kwargs)


class LatticeObservable(ElementObservable):
    """Observe an attribute of selected lattice elements."""

    def __init__(
        self,
        refpts: Refpts,
        attrname: str,
        index: int | None = None,
        name: str | None = None,
        **kwargs,
    ):
        # noinspection PyUnresolvedReferences
        r"""Args:
            refpts:         Elements to be observed
              See ":ref:`Selecting elements in a lattice <refpts>`"
            attrname:       Attribute name
            index:          Index in the attribute array. If :py:obj:`None`,
              the whole array is specified
            name:           Observable name. If :py:obj:`None`, an explicit
              name will be generated.

        Keyword Args:
            postfun:        Post-processing function. It can be any numpy ufunc or a
              function name in {"real", "imag", "abs", "angle", "log", "exp", "sqrt"}.
            statfun:        Statistics post-processing function. it can be a numpy
              function or a function name in {"mean", "std", "var", "min", "max"}.
              Example: :pycode:`statfun=numpy.mean`.

        Example:

            >>> obs = LatticeObservable(
            ...     at.Sextupole, "KickAngle", index=0, statfun=np.sum
            ... )

            Observe the sum of horizontal kicks in Sextupoles
        """
        fun = _Ring(attrname, index, refpts)
        needs = {Need.RING}
        name = self._set_name(name, attrname, index)
        super().__init__(fun, refpts, needs=needs, name=name, **kwargs)


class TrajectoryObservable(ElementObservable):
    """Observe trajectory coordinates at selected locations.

    Process the *r_out* output if :py:meth:`.Lattice.track`
    """

    def __init__(
        self,
        refpts: Refpts,
        axis: AxisDef = Ellipsis,
        name: str | None = None,
        **kwargs,
    ):
        r"""Args:
            refpts:         Observation points.
              See ":ref:`Selecting elements in a lattice <refpts>`"
            axis:          Index in the orbit array, If :py:obj:`Ellipsis`,
              the whole array is specified
            name:           Observable name. If :py:obj:`None`, an explicit
              name will be generated.

        Keyword Args:
            postfun:        Post-processing function. It can be any numpy ufunc or a
               function name in {"real", "imag", "abs", "angle", "log", "exp", "sqrt"}.
            statfun:        Statistics post-processing function. it can be a numpy
              function or a function name in {"mean", "std", "var", "min", "max"}.
              Example: :pycode:`statfun=numpy.mean`.
            target:         Target value for a constraint. If :py:obj:`None`
              (default), the residual will always be zero.
            weight:         Weight factor: the residual is
              :pycode:`((value-target)/weight)**2`
            bounds:         Tuple of lower and upper bounds. The parameter
              is constrained in the interval
              [*target*\ +\ *low_bound* *target*\ +\ *up_bound*]

        The *target*, *weight* and *bounds* inputs must be broadcastable to the
        shape of *value*.
        """
        name = self._set_name(name, "trajectory", axis_(axis, key="code"))
        fun = _ArrayAccess(axis_(axis, key="index"))
        needs = {Need.TRAJECTORY}
        super().__init__(fun, refpts, needs=needs, name=name, **kwargs)


class EmittanceObservable(Observable):
    """Observe emittance-related parameters.

    Process the output of :py:func:`.envelope_parameters`.
    """

    def __init__(
        self, param: str, plane: AxisDef = None, name: str | None = None, **kwargs
    ):
        r"""Args:
            param:          Parameter name (see :py:func:`.envelope_parameters`) or
              :ref:`user-defined evaluation function <emittance_eval>`
            plane:          One out of {0, 'x', 'h', 'H'} for horizontal plane,
             one out of {1, 'y', 'v', 'V'} for vertival plane or one out of
             {2, 'z', 'l', 'L'} for longitudinal plane
            name:           Observable name. If :py:obj:`None`, an explicit
              name will be generated.

        Keyword Args:
            postfun:        Post-processing function. It can be any numpy ufunc or a
              function name in {"real", "imag", "abs", "angle", "log", "exp", "sqrt"}.
            statfun:        Statistics post-processing function. it can be a numpy
              function or a function name in {"mean", "std", "var", "min", "max"}.
              Example: :pycode:`statfun=numpy.mean`.
            target:         Target value for a constraint. If :py:obj:`None`
              (default), the residual will always be zero.
            weight:         Weight factor: the residual is
              :pycode:`((value-target)/weight)**2`
            bounds:         Tuple of lower and upper bounds. The parameter
              is constrained in the interval
              [*target*\ +\ *low_bound* *target*\ +\ *up_bound*]

        .. _emittance_eval:
        .. rubric:: User-defined evaluation function

        It is called as:

        :pycode:`value = fun(paramdata)`

        *paramdata* if the :py:class:`.RingParameters` object returned by
        :py:func:`.envelope_parameters`.

        *value* is the value of the Observable.

        Example:

            >>> EmittanceObservable("emittances", plane="h")

            Observe the horizontal emittance
        """
        name = self._set_name(name, param, plane_(plane, key="code"))
        if callable(param):
            fun = param
        else:
            fun = partial(_record_access, param, plane_(plane, key="index"))
        needs = {Need.EMITTANCE}
        super().__init__(fun, needs=needs, name=name, **kwargs)


# noinspection PyPep8Naming
def GlobalOpticsObservable(
    param: str,
    plane: AxisDef = Ellipsis,
    name: str | None = None,
    use_integer: bool = False,
    **kwargs,
):
    # noinspection PyUnresolvedReferences
    r"""Observe a global optics parameter.

    Process the *ringdata* output of :py:func:`.get_optics`.

    Args:
        param:          Optics parameter name (see :py:func:`.get_optics`)
          or :ref:`user-defined evaluation function <globaloptics_eval>`
        plane:          Index in the parameter array, If :py:obj:`Ellipsis`,
          the whole array is specified
        name:           Observable name. If :py:obj:`None`, an explicit
          name will be generated
        use_integer:    For *'tune'* parameter: derive the tune from the
          phase advance to avoid skipping integers. Slower than looking only
          at the fractional part

    Keyword Args:
        target:         Target value for a constraint. If :py:obj:`None`
          (default), the residual will always be zero.
        weight:         Weight factor: the residual is
          :pycode:`((value-target)/weight)**2`
        bounds:         Tuple of lower and upper bounds. The parameter
          is constrained in the interval
          [*target*\ +\ *low_bound* *target*\ +\ *up_bound*]
        postfun:        Post-processing function. It can be any numpy ufunc or a
              function name in {"real", "imag", "abs", "angle", "log", "exp", "sqrt"}.

    The *target*, *weight* and *bounds* inputs must be broadcastable to the
    shape of *value*.

    .. _globaloptics_eval:
    .. rubric:: User-defined evaluation function

    It is called as:

    :pycode:`value = fun(ring, ringdata)`

    - *ringdata* is the output of :py:func:`.get_optics`,
    - *value* is the value of the Observable.

    Examples:

        >>> obs = GlobalOpticsObservable("tune", use_integer=True)

        Observe the tune in both planes, including the integer part (slower)

        >>> obs = GlobalOpticsObservable("chromaticity", plane="v")

        Observe the vertical chromaticity
    """
    if param == "tune" and use_integer:
        # noinspection PyProtectedMember
        name = ElementObservable._set_name(name, param, plane_(plane, key="code"))
        return LocalOpticsObservable(
            End,
            _Tune(plane_(plane, key="index")),
            name=name,
            summary=True,
            all_points=True,
            **kwargs,
        )
    else:
        return _GlobalOpticsObservable(param, plane=plane, name=name, **kwargs)



r"""
Helper functions for working with AT lattices.

A lattice as understood by pyAT is a sequence of elements.  These functions
are useful for working with these sequences.

.. _refpts:

**Selecting elements in a lattice:**

The *refpts* argument allows functions to select locations in the lattice. The
location is defined as the entrance of the selected element. *refpts* may be:

#. an integer in the range *[-len(lattice), len(lattice)-1]*
   selecting an element according to the python indexing rules.
   As a special case, *len(lattice)* is allowed and refers to the end
   of the last element,
#. an ordered list of such integers without duplicates,
#. a numpy array of booleans of maximum length *len(lattice)+1*,
   where selected elements are :py:obj:`True`,
#. :py:obj:`None`, meaning an empty selection,
#. :py:obj:`.All`, meaning "all possible reference points": the entrance of all
   elements plus the end of the last element,
#. :py:obj:`.End`, selecting the end of the last element,
#. an element type, selecting all the elements of that type in
   the lattice, e.g. :pycode:`at.Sextupole`,
#. a string, selecting all the elements whose `FamName` attribute matches it.
   Unix shell-style wildcards are accepted, e.g. `"Q[FD]*"`,
#. a callable :pycode:`filtfunc` such that :pycode:`filtfunc(elem)`
   is :py:obj:`True` for selected elements.

"""

from __future__ import annotations

import functools
import re
from enum import Enum
from fnmatch import fnmatch
from itertools import compress
from operator import attrgetter
from typing import Union

# Necessary for type aliases in python <= 3.8 :
# from collections.abc import Callable, Sequence, Iterator
from typing import Callable, Sequence, Iterator, Type

import numpy
import numpy.typing as npt

from .elements import Element, Dipole

_GEOMETRY_EPSIL = 1.0e-3

__all__ = [
    "All",
    "End",
    "AtError",
    "AtWarning",
    "BoolRefpts",
    "Uint32Refpts",
    "check_radiation",
    "check_6d",
    "set_radiation",
    "set_6d",
    "make_copy",
    "uint32_refpts",
    "bool_refpts",
    "get_uint32_index",
    "get_bool_index",
    "checkattr",
    "checktype",
    "checkname",
    "get_elements",
    "get_s_pos",
    "refpts_count",
    "refpts_iterator",
    "get_value_refpts",
    "set_value_refpts",
    "Refpts",
    "get_geometry",
    "setval",
    "getval",
]


class AtError(Exception):
    pass


class AtWarning(UserWarning):
    pass


_typ1 = "None, All, End, int, bool"

_typ2 = "None, All, End, int, bool, str, type[Element], ElementFilter"


class RefptsCode(Enum):
    All = "All"
    End = "End"


ElementFilter = Callable[[Element], bool]
BoolRefpts = npt.NDArray[bool]
Uint32Refpts = npt.NDArray[numpy.uint32]
RefIndex = Union[None, int, Sequence[int], bool, Sequence[bool], RefptsCode]
Refpts = Union[Type[Element], Element, ElementFilter, str, RefIndex]


#: :py:obj:`All` is a special value to be used as *refpts*. It means
#: "all possible reference points": the entrance of all elements plus the end
#: of the last element.
All = RefptsCode.All

#: :py:obj:`End` is a special value to be used as *refpts*. It refers to the
#: end of the last element.
End = RefptsCode.End


def _chkattr(attrname: str, el):
    return hasattr(el, attrname)


def _chkattrval(attrname: str, attrvalue, el):
    try:
        v = getattr(el, attrname)
    except AttributeError:
        return False
    else:
        return v == attrvalue


def _chkpattern(pattern: str, el):
    return fnmatch(el.FamName, pattern)


def _chkregex(pattern: str, el):
    rgx = re.compile(pattern)
    return rgx.fullmatch(el.FamName)


def _chktype(eltype: type, el):
    return isinstance(el, eltype)


def _type_error(refpts, types):
    if isinstance(refpts, numpy.ndarray):
        tp = refpts.dtype.type
    else:
        tp = type(refpts)
    return TypeError(f"Invalid refpts type {tp}. Allowed types: {types}")


# setval and getval return pickleable functions: no inner, nested function
# is allowed. So nested functions are replaced be module-level callable
# class instances
class _AttrItemGetter:
    __slots__ = ["attrname", "index"]

    def __init__(self, attrname: str, index: int):
        self.attrname = attrname
        self.index = index

    def __call__(self, elem):
        return getattr(elem, self.attrname)[self.index]


def getval(attrname: str, index: int | None = None) -> Callable:
    """Return a callable object which fetches item *index* of
    attribute *attrname* of its operand. Examples:

    - After ``f = getval('Length')``, ``f(elem)`` returns ``elem.Length``
    - After ``f = getval('PolynomB', index=1)``, ``f(elem)`` returns
      ``elem.PolynomB[1]``

    """
    if index is None:
        return attrgetter(attrname)
    else:
        return _AttrItemGetter(attrname, index)


class _AttrSetter:
    __slots__ = ["attrname"]

    def __init__(self, attrname: str):
        self.attrname = attrname

    def __call__(self, elem, value):
        setattr(elem, self.attrname, value)


class _AttrItemSetter:
    __slots__ = ["attrname", "index"]

    def __init__(self, attrname: str, index: int):
        self.attrname = attrname
        self.index = index

    def __call__(self, elem, value):
        getattr(elem, self.attrname)[self.index] = value


def setval(attrname: str, index: int | None = None) -> Callable:
    """Return a callable object which sets the value of  item *index* of
    attribute *attrname* of its 1st argument to it 2nd orgument.

    - After ``f = setval('Length')``, ``f(elem, value)`` is equivalent to
      ``elem.Length = value``
    - After ``f = setval('PolynomB', index=1)``, ``f(elem, value)`` is
      equivalent to ``elem.PolynomB[1] = value``

    """
    if index is None:
        return _AttrSetter(attrname)
    else:
        return _AttrItemSetter(attrname, index)


def check_radiation(rad: bool) -> Callable:
    r"""Deprecated decorator for optics functions (see :py:func:`check_6d`).

    :meta private:
    """
    return check_6d(rad)


def set_radiation(rad: bool) -> Callable:
    r"""Deprecated decorator for optics functions (see :py:func:`set_6d`).

    :meta private:
    """
    return set_6d(rad)


def check_6d(is_6d: bool) -> Callable:
    r"""Decorator for optics functions

    Wraps a function like :pycode:`func(ring, *args, **kwargs)` where
    :pycode:`ring` is a :py:class:`.Lattice` object. Raise :py:class:`.AtError`
    if :pycode:`ring.is_6d` is not the desired *is_6d* state. No test is
    performed if :pycode:`ring` is not a :py:class:`.Lattice`.

    Parameters:
        is_6d: Desired 6D state

    Raises:
        AtError: if :pycode:`ring.is_6d` is not *is_6d*

    Example:

        >>> @check_6d(False)
        ... def find_orbit4(ring, dct=None, guess=None, **kwargs):
                ...

        Raises :py:class:`.AtError` if :pycode:`ring.is_6d` is :py:obj:`True`

    See Also:
        :py:func:`set_6d`
    """

    def radiation_decorator(func):
        @functools.wraps(func)
        def wrapper(ring, *args, **kwargs):
            ringrad = getattr(ring, "is_6d", is_6d)
            if ringrad != is_6d:
                raise AtError(f'{func.__name__} needs "ring.is_6d" {is_6d}')
            return func(ring, *args, **kwargs)

        return wrapper

    return radiation_decorator


def set_6d(is_6d: bool) -> Callable:
    # noinspection PyUnresolvedReferences
    r"""Decorator for optics functions

    Wraps a function like :pycode:`func(ring, *args, **kwargs)` where
    :pycode:`ring` is a :py:class:`.Lattice` object. If :pycode:`ring.is_6d`
    is not the desired *is_6d* state, :pycode:`func()` will be called with a
    modified copy of :pycode:`ring` satisfying the *is_6d* state.

    Parameters:
        is_6d: Desired 6D state

    Example:

        >>> @set_6d(True)
        ... def compute(ring)
                ...

        is roughly equivalent to:

        >>> if ring.is_6d:
        ...     compute(ring)
        ... else:
        ...     compute(ring.enable_6d(copy=True))

        See Also:
            :py:func:`check_6d`, :py:meth:`.Lattice.enable_6d`,
            :py:meth:`.Lattice.disable_6d`
    """
    if is_6d:

        def setrad_decorator(func):
            @functools.wraps(func)
            def wrapper(ring, *args, **kwargs):
                rg = ring if ring.is_6d else ring.enable_6d(copy=True)
                return func(rg, *args, **kwargs)

            return wrapper

    else:

        def setrad_decorator(func):
            @functools.wraps(func)
            def wrapper(ring, *args, **kwargs):
                rg = ring.disable_6d(copy=True) if ring.is_6d else ring
                return func(rg, *args, **kwargs)

            return wrapper

    return setrad_decorator


def make_copy(copy: bool) -> Callable:
    r"""Decorator for optics functions

    Wraps a function like :pycode:`func(ring, refpts, *args, **kwargs)` where
    :pycode:`ring` is a :py:class:`.Lattice` object.

    If *copy* is :py:obj:`False`, the function is not modified,

    If *copy* is :py:obj:`True`, a shallow copy of :pycode:`ring` is done,
    then the elements selected by *refpts* are deep-copied, then :pycode:`func`
    is applied to the copy, and the new :pycode:`ring` is returned.

    Parameters:
        copy:   If :py:obj:`True`, apply the decorated function to a copy of
          :pycode:`ring`
    """
    if copy:

        def copy_decorator(func):
            @functools.wraps(func)
            def wrapper(ring, refpts, *args, **kwargs):
                try:
                    ring = ring.replace(refpts)
                except AttributeError:
                    check = get_bool_index(ring, refpts)
                    ring = [el.deepcopy() if ok else el for el, ok in zip(ring, check)]
                func(ring, refpts, *args, **kwargs)
                return ring

            return wrapper

    else:

        def copy_decorator(func):
            return func

    return copy_decorator


def uint32_refpts(
    refpts: RefIndex, n_elements: int, endpoint: bool = True, types: str = _typ1
) -> Uint32Refpts:
    r"""Return a :py:obj:`~numpy.uint32` array of element indices selecting
    ring elements.

    Parameters:
        refpts:     Element selector. *refpts* may be:

          #. an integer or a sequence of integers
             (0 indicating the first element),
          #. a sequence of booleans marking the selected elements,
          #. :py:obj:`None`, meaning empty selection,
          #. :py:obj:`.All`, meaning "all possible reference points",
          #. :py:obj:`.End`, selecting the end of the last element.
        n_elements: Length of the sequence of elements
        endpoint:   if :py:obj:`True`, allow *n_elements* as a
          special index, referring to the end of the last element.
        types:      Allowed types

    Returns:
        uint32_ref (Uint32Refpts):  :py:obj:`~numpy.uint32` numpy array used
          for indexing :py:class:`.Element`\ s in a lattice.
    """
    refs = numpy.ravel(refpts)
    if refpts is RefptsCode.All:
        stop = n_elements + 1 if endpoint else n_elements
        return numpy.arange(stop, dtype=numpy.uint32)
    elif refpts is RefptsCode.End:
        if not endpoint:
            raise IndexError('"End" index out of range')
        return numpy.array([n_elements], dtype=numpy.uint32)
    elif (refpts is None) or (refs.size == 0):
        return numpy.array([], dtype=numpy.uint32)
    elif numpy.issubdtype(refs.dtype, numpy.bool_):
        return numpy.flatnonzero(refs).astype(numpy.uint32)
    elif numpy.issubdtype(refs.dtype, numpy.integer):
        # Handle negative indices
        if endpoint:
            refs = numpy.array(
                [i if (i == n_elements) else i % n_elements for i in refs],
                dtype=numpy.uint32,
            )
        else:
            refs = numpy.array([i % n_elements for i in refs], dtype=numpy.uint32)
        # Check ascending
        if refs.size > 1:
            prev = refs[0]
            for nxt in refs[1:]:
                if nxt < prev:
                    raise IndexError("Index out of range or not in ascending order")
                elif nxt == prev:
                    raise IndexError("Duplicated index")
                prev = nxt

        return refs
    else:
        raise _type_error(refpts, types)


# noinspection PyIncorrectDocstring
def get_uint32_index(
    ring: Sequence[Element], refpts: Refpts, endpoint: bool = True, regex: bool = False
) -> Uint32Refpts:
    # noinspection PyUnresolvedReferences, PyShadowingNames
    r"""Returns an integer array of element indices, selecting ring elements.

    Parameters:
        refpts:         Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        endpoint:   if :py:obj:`True`, allow *len(ring)* as a
          special index, referring to the end of the last element.
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        uint32_ref (Uint32Refpts): uint32 numpy array used for indexing
          :py:class:`.Element`\ s in a lattice.

    Examples:

        >>> get_uint32_index(ring, at.Sextupole)
        array([ 21,  27,  35,  89,  97, 103], dtype=uint32)

        numpy array of indices of all :py:class:`.Sextupole`\ s

        >>> get_uint32_index(ring, at.End)
        array([121], dtype=uint32)

        numpy array([:pycode:`len(ring)+1`])

        >>> get_uint32_index(ring, at.checkattr("Frequency"))
        array([0], dtype=uint32)

        numpy array of indices of all elements having a 'Frequency'
        attribute
    """
    if isinstance(refpts, type):
        checkfun = checktype(refpts)
    elif callable(refpts):
        checkfun = refpts
    elif isinstance(refpts, Element):
        checkfun = checktype(type(refpts))
    elif isinstance(refpts, str):
        checkfun = checkname(refpts, regex=regex)
    else:
        return uint32_refpts(refpts, len(ring), endpoint=endpoint, types=_typ2)

    return numpy.fromiter(
        (i for i, el in enumerate(ring) if checkfun(el)), dtype=numpy.uint32
    )


def bool_refpts(
    refpts: RefIndex, n_elements: int, endpoint: bool = True, types: str = _typ1
) -> BoolRefpts:
    r"""Returns a :py:class:`bool` array of element indices, selecting ring
    elements.

    Parameters:
        refpts:     Element selector. *refpts* may be:

          #. an integer or a sequence of integers
             (0 indicating the first element),
          #. a sequence of booleans marking the selected elements,
          #. :py:obj:`None`, meaning empty selection,
          #. :py:obj:`.All`, meaning "all possible reference points",
          #. :py:obj:`.End`, selecting the end of the last element.
        n_elements: Length of the lattice
        endpoint:   if :py:obj:`True`, allow *n_elements* as a
          special index, referring to the end of the last element.
        types:      Allowed types

    Returns:
        bool_refs (BoolRefpts):  A bool numpy array used for indexing
          :py:class:`.Element`\ s in a lattice.
    """
    refs = numpy.ravel(refpts)
    stop = n_elements + 1 if endpoint else n_elements
    if refpts is RefptsCode.All:
        return numpy.ones(stop, dtype=bool)
    elif refpts is RefptsCode.End:
        if not endpoint:
            raise IndexError('"End" index out of range')
        brefpts = numpy.zeros(stop, dtype=bool)
        brefpts[n_elements] = True
        return brefpts
    elif (refpts is None) or (refs.size == 0):
        return numpy.zeros(stop, dtype=bool)
    elif numpy.issubdtype(refs.dtype, numpy.bool_):
        diff = stop - refs.size
        if diff <= 0:
            return refs[:stop]
        else:
            return numpy.append(refs, numpy.zeros(diff, dtype=bool))
    elif numpy.issubdtype(refs.dtype, numpy.integer):
        brefpts = numpy.zeros(stop, dtype=bool)
        brefpts[refs] = True
        return brefpts
    else:
        raise _type_error(refpts, types)


# noinspection PyIncorrectDocstring
def get_bool_index(
    ring: Sequence[Element], refpts: Refpts, endpoint: bool = True, regex: bool = False
) -> BoolRefpts:
    # noinspection PyUnresolvedReferences, PyShadowingNames
    r"""Returns a bool array of element indices, selecting ring elements.

    Parameters:
        refpts:         Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        endpoint:   if :py:obj:`True`, allow *len(ring)* as a
          special index, referring to the end of the last element.
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        bool_refs (BoolRefpts):  A bool numpy array used for indexing
          :py:class:`.Element`\ s in a lattice.

    Examples:

        >>> refpts = get_bool_index(ring, at.Quadrupole)

        Returns a numpy array of booleans where all :py:class:`.Quadrupole`
        are :py:obj:`True`

        >>> refpts = get_bool_index(ring, "Q[FD]*")

        Returns a numpy array of booleans where all elements whose *FamName*
        matches "Q[FD]*" are :py:obj:`True`

        >>> refpts = get_bool_index(ring, at.checkattr("K", 0.0))

        Returns a numpy array of booleans where all elements whose *K*
        attribute is 0.0 are :py:obj:`True`

        >>> refpts = get_bool_index(ring, None)

        Returns a numpy array of *len(ring)+1* :py:obj:`False` values
    """
    if isinstance(refpts, type):
        checkfun = checktype(refpts)
    elif callable(refpts):
        checkfun = refpts
    elif isinstance(refpts, Element):
        checkfun = checktype(type(refpts))
    elif isinstance(refpts, str):
        checkfun = checkname(refpts, regex=regex)
    else:
        return bool_refpts(refpts, len(ring), endpoint=endpoint, types=_typ2)

    boolrefs = numpy.fromiter(map(checkfun, ring), dtype=bool, count=len(ring))
    if endpoint:
        boolrefs = numpy.append(boolrefs, False)
    return boolrefs


def checkattr(attrname: str, attrvalue=None) -> ElementFilter:
    # noinspection PyUnresolvedReferences
    r"""Checks the presence or the value of an attribute

    Returns a function to be used as an :py:class:`.Element` filter, which
    checks the presence or the value of an attribute of the
    provided :py:class:`.Element`.
    This function can be used to extract from a ring all elements
    having a given attribute.

    Parameters:
        attrname: Attribute name
        attrvalue: Attribute value. If absent, the returned function checks
          the presence of an *attrname* attribute. If present, the
          returned function checks if :pycode:`attrname == attrvalue`.

    Returns:
        checkfun (ElementFilter):   Element filter function

    Examples:

        >>> cavs = filter(checkattr("Frequency"), ring)

        Returns an iterator over all elements in *ring* that have a
        :pycode:`Frequency` attribute

        >>> elts = filter(checkattr("K", 0.0), ring)

        Returns an iterator over all elements of ring that have a
        :pycode:`K` attribute equal to 0.0
    """
    if attrvalue is None:
        return functools.partial(_chkattr, attrname)
    else:
        return functools.partial(_chkattrval, attrname, attrvalue)


def checktype(eltype: type | tuple[type, ...]) -> ElementFilter:
    # noinspection PyUnresolvedReferences
    r"""Checks the type of an element

    Returns a function to be used as an :py:class:`.Element` filter, which
    checks the type of the provided :py:class:`.Element`.
    This function can be used to extract from a ring all elements
    having a given type.

    Parameters:
        eltype: Desired :py:class:`.Element` type

    Returns:
        checkfun (ElementFilter):   Element filter function

    Examples:

        >>> qps = filter(checktype(at.Quadrupole), ring)

        Returns an iterator over all quadrupoles in ring
    """
    return functools.partial(_chktype, eltype)


def checkname(pattern: str, regex: bool = False) -> ElementFilter:
    # noinspection PyUnresolvedReferences
    r"""Checks the name of an element

    Returns a function to be used as an :py:class:`.Element` filter,
    which checks the name of the provided :py:class:`.Element`.
    This function can be used to extract from a ring all elements
    having a given name.

    Parameters:
        pattern: Desired :py:class:`.Element` name. Unix shell-style
          wildcards are supported (see :py:func:`fnmatch.fnmatch`)
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        checkfun (ElementFilter):   Element filter function

    Examples:

        >>> qps = filter(checkname("QF*"), ring)

        Returns an iterator over all the elements with name starting with ``QF``.
    """
    if regex:
        return functools.partial(_chkregex, pattern)
    else:
        return functools.partial(_chkpattern, pattern)


def refpts_iterator(
    ring: Sequence[Element], refpts: Refpts, regex: bool = False
) -> Iterator[Element]:
    r"""Return an iterator over selected elements in a lattice

    Parameters:
        ring:           Lattice description
        refpts:         Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        elem_iter (Iterator[Element]):  Iterator over the elements in *ring*
          selected by *refpts*.
    """
    if isinstance(refpts, type):
        checkfun = checktype(refpts)
    elif callable(refpts):
        checkfun = refpts
    elif isinstance(refpts, Element):
        checkfun = checktype(type(refpts))
    elif isinstance(refpts, str):
        checkfun = checkname(refpts, regex=regex)
    else:
        refs = numpy.ravel(refpts)
        if refpts is RefptsCode.All:
            return (el for el in ring)
        elif refpts is RefptsCode.End:
            raise IndexError('"End" is not allowed for endpoint=False')
        elif (refpts is None) or (refs.size == 0):
            return iter(())
        elif numpy.issubdtype(refs.dtype, numpy.bool_):
            return compress(ring, refs)
        elif numpy.issubdtype(refs.dtype, numpy.integer):
            return (ring[i] for i in refs)
        else:
            raise _type_error(refpts, _typ2)

    return filter(checkfun, ring)


# noinspection PyUnusedLocal,PyIncorrectDocstring
def refpts_count(
    refpts: RefIndex, n_elements: int, endpoint: bool = True, types: str = _typ1
) -> int:
    r"""Returns the number of reference points

    Parameters:
        refpts:     refpts may be:

          #. an integer or a sequence of integers
             (0 indicating the first element),
          #. a sequence of booleans marking the selected elements,
          #. :py:obj:`None`, meaning empty selection,
          #. :py:obj:`.All`, meaning "all possible reference points",
          #. :py:obj:`.End`, selecting the end of the last element.
        n_elements: Lattice length
        endpoint:   if :py:obj:`True`, allow *n_elements* as a
          special index, referring to the end of the last element.

    Returns:
        nrefs (int):  The number of reference points
    """
    refs = numpy.ravel(refpts)
    if refpts is RefptsCode.All:
        return n_elements + 1 if endpoint else n_elements
    elif refpts is RefptsCode.End:
        if not endpoint:
            raise IndexError('"End" index out of range')
        return 1
    elif (refpts is None) or (refs.size == 0):
        return 0
    elif numpy.issubdtype(refs.dtype, numpy.bool_):
        return numpy.count_nonzero(refs)
    elif numpy.issubdtype(refs.dtype, numpy.integer):
        return len(refs)
    else:
        raise _type_error(refpts, types)


def _refcount(
    ring: Sequence[Element], refpts: Refpts, endpoint: bool = True, regex: bool = False
) -> int:
    # noinspection PyUnresolvedReferences, PyShadowingNames
    r"""Returns the number of reference points

    Parameters:
        refpts:         Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        endpoint:   if :py:obj:`True`, allow *len(ring)* as a
          special index, referring to the end of the last element.
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        nrefs (int):  The number of reference points

    Examples:

        >>> refpts = ring.refcount(at.Sextupole)
        6

        Returns the number of :py:class:`.Sextupole`\ s in the lattice

        >>> refpts = ring.refcount(at.All)
        122

        Returns *len(ring)+1*

        >>> refpts = ring.refcount(at.All, endpoint=False)
        121

        Returns *len(ring)*
    """
    if isinstance(refpts, type):
        checkfun = checktype(refpts)
    elif callable(refpts):
        checkfun = refpts
    elif isinstance(refpts, Element):
        checkfun = checktype(type(refpts))
    elif isinstance(refpts, str):
        checkfun = checkname(refpts, regex=regex)
    else:
        return refpts_count(refpts, len(ring), endpoint=endpoint, types=_typ2)

    return len(list(filter(checkfun, ring)))


# noinspection PyUnusedLocal,PyIncorrectDocstring
def get_elements(ring: Sequence[Element], refpts: Refpts, regex: bool = False) -> list:
    r"""Returns a list of elements selected by *key*.

    Deprecated: :pycode:`get_elements(ring, refpts)` is :pycode:`ring[refpts]`

    Parameters:
        ring:           Lattice description
        refpts:         Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        elem_list (list):  list of :py:class:`.Element`\ s matching key
    """
    return list(refpts_iterator(ring, refpts, regex=regex))


def get_value_refpts(
    ring: Sequence[Element],
    refpts: Refpts,
    attrname: str,
    index: int | None = None,
    regex: bool = False,
):
    r"""Extracts attribute values from selected
        lattice :py:class:`.Element`\ s.

    Parameters:
        ring:           Lattice description
        refpts:         Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        attrname:   Attribute name
        index:      index of the value to retrieve if *attrname* is
          an array. If :py:obj:`None` the full array is retrieved
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        attrvalues: numpy Array of attribute values.
    """
    getf = getval(attrname, index=index)
    return numpy.array(
        [getf(elem) for elem in refpts_iterator(ring, refpts, regex=regex)]
    )


def set_value_refpts(
    ring: Sequence[Element],
    refpts: Refpts,
    attrname: str,
    attrvalues,
    index: int | None = None,
    increment: bool = False,
    copy: bool = False,
    regex: bool = False,
):
    r"""Set the values of an attribute of an array of elements based on
    their refpts


    Parameters:
        ring:       Lattice description
        refpts:         Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        attrname:   Attribute name
        attrvalues: Attribute values
        index:      index of the value to set if *attrname* is
          an array. if :py:obj:`None`, the full array is replaced by
          *attrvalue*
        increment:  If :py:obj:`True`, *attrvalues* are added to the
          initial values.
        regex: Use regular expression for *refpts* string matching
          instead of Unix shell-style wildcards.
        copy:       If :py:obj:`False`, the modification is done in-place,

          If :py:obj:`True`, return a shallow copy of the lattice. Only the
          modified elements are copied.

          .. Caution::

             a shallow copy means that all non-modified
             elements are shared with the original lattice.
             Any further modification will affect both lattices.
    """
    setf = setval(attrname, index=index)
    if increment:
        attrvalues += get_value_refpts(ring, refpts, attrname, index=index, regex=regex)
    else:
        attrvalues = numpy.broadcast_to(
            attrvalues, (_refcount(ring, refpts, regex=regex),)
        )

    # noinspection PyShadowingNames
    @make_copy(copy)
    def apply(ring, refpts, values, regex):
        for elm, val in zip(refpts_iterator(ring, refpts, regex=regex), values):
            setf(elm, val)

    return apply(ring, refpts, attrvalues, regex)


def get_s_pos(
    ring: Sequence[Element], refpts: Refpts = All, regex: bool = False
) -> Sequence[float]:
    # noinspection PyUnresolvedReferences
    r"""Returns the locations of selected elements

    Parameters:
        ring:       Lattice description
        refpts:     Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        s_pos:  Array of locations of the elements selected by *refpts*

    Example:

        >>> get_s_pos(ring, at.End)
        array([26.37428795])

        Position at the end of the last element: length of the lattice
    """
    # Positions at the end of each element.
    s_pos = numpy.cumsum([getattr(el, "Length", 0.0) for el in ring])
    # Prepend position at the start of the first element.
    s_pos = numpy.concatenate(([0.0], s_pos))
    return s_pos[get_bool_index(ring, refpts, regex=regex)]


def get_geometry(
    ring: list[Element],
    refpts: Refpts = All,
    start_coordinates: tuple[float, float, float] = (0, 0, 0),
    centered: bool = False,
    regex: bool = False,
):
    # noinspection PyShadowingNames
    r"""Compute the 2D ring geometry in cartesian coordinates

    Parameters:
        ring:               Lattice description.
        refpts:     Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        start_coordinates:  *x*, *y*, *angle* at starting point. *x*
          and *y* are ignored if *centered* is :py:obj:`True`.
        centered:           if :py:obj:`True` the coordinates origin is the
          centre of the ring.
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        geomdata:           recarray containing, x, y, angle.
        radius:             machine radius at the beginning of the lattice.

            .. attention::
               This radius is different from the radius usually defined as
               :math:`C/2\pi`

    Example:

       >>> geomdata, radius = get_geometry(ring)
    """

    geom_dtype = [("x", numpy.float64), ("y", numpy.float64), ("angle", numpy.float64)]
    boolrefs = get_bool_index(ring, refpts, endpoint=True, regex=regex)
    nrefs = refpts_count(boolrefs, len(ring))
    geomdata = numpy.recarray((nrefs,), dtype=geom_dtype)
    xx = numpy.zeros(len(ring) + 1)
    yy = numpy.zeros(len(ring) + 1)
    angle = numpy.zeros(len(ring) + 1)
    x0, y0, t0 = start_coordinates
    x, y = 0.0, 0.0
    t = t0

    xx[0] = x
    yy[0] = y
    angle[0] = t
    for ind, el in enumerate(ring):
        ll = el.Length
        if isinstance(el, Dipole) and el.BendingAngle != 0:
            ang = 0.5 * el.BendingAngle
            ll *= numpy.sin(ang) / ang
        else:
            ang = 0.0
        t -= ang
        x += ll * numpy.cos(t)
        y += ll * numpy.sin(t)
        t -= ang
        xx[ind + 1] = x
        yy[ind + 1] = y
        angle[ind + 1] = t

    dff = (t + _GEOMETRY_EPSIL) % (2.0 * numpy.pi) - _GEOMETRY_EPSIL
    if abs(dff) < _GEOMETRY_EPSIL:
        xcenter = numpy.mean(xx)
        ycenter = numpy.mean(yy)
    elif abs(dff - numpy.pi) < _GEOMETRY_EPSIL:
        xcenter = 0.5 * x
        ycenter = 0.5 * y
    else:
        num = numpy.cos(t) * x + numpy.sin(t) * y
        den = numpy.sin(t - t0)
        xcenter = -num * numpy.sin(t0) / den
        ycenter = num * numpy.cos(t0) / den
    radius = numpy.sqrt(xcenter * xcenter + ycenter * ycenter)
    if centered:
        xx -= xcenter
        yy -= ycenter
    else:
        xx += x0
        yy += y0
    geomdata["x"] = xx[boolrefs]
    geomdata["y"] = yy[boolrefs]
    geomdata["angle"] = angle[boolrefs]
    return geomdata, radius



"""AT plotting functions"""

from __future__ import annotations

from math import sqrt

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes

from at.constants import clight
from at.lattice import Lattice, axis_
from at.lattice import RFCavity
from at.physics import get_mcf


# Function to compute and plot acceptance
def plot_acceptance(ring: Lattice, planes, *args, **kwargs):
    # noinspection PyUnresolvedReferences
    r"""Plots the acceptance

    Computes the acceptance at repfts observation points using
    :py:func:`.get_acceptance` and plots the tracked
    and survived particles, and the acceptance boundary.

    Parameters:
        ring:           Lattice definition
        planes:         max. dimension 2, Plane(s) to scan for the acceptance.
          Allowed values are: *'x'*, *'xp'*, *'y'*,
          *'yp'*, *'dp'*, *'ct'*

    Keyword Args:
        acceptance (tuple[ndarray, ndarray, ndarray]): tuple containing
          pre-computed acceptance *(boundary, survived, grid)*
        npoints:        (len(planes),) array: number of points in each
          dimension
        amplitudes:     (len(planes),) array: set the search range:

          * :py:attr:`GridMode.CARTESIAN/RADIAL <.GridMode.RADIAL>`:
            max. amplitude
          * :py:attr:`.GridMode.RECURSIVE`: initial step
        nturns (int):       Number of turns for the tracking
        obspt (Refpts):    Observation points. Default: start of the machine
        dp (float):         Static momentum offset
        offset:             Initial orbit. Default: closed orbit
        bounds:             Defines the tracked range: range=bounds*amplitude.
          It can be used to select quadrants. For example, default values are:

          * :py:attr:`.GridMode.CARTESIAN`: ((-1, 1), (0, 1))
          * :py:attr:`GridMode.RADIAL/RECURSIVE <.GridMode.RADIAL>`: ((0, 1),
            (:math:`\pi`, 0))
        grid_mode (GridMode):   Defines the evaluation grid:

          * :py:attr:`.GridMode.CARTESIAN`: full [:math:`\:x, y\:`] grid
          * :py:attr:`.GridMode.RADIAL`: full [:math:`\:r, \theta\:`] grid
          * :py:attr:`.GridMode.RECURSIVE`: radial recursive search
        use_mp (bool):      Use python multiprocessing (:py:func:`.patpass`,
          default use :py:func:`.lattice_pass`). In case multiprocessing is not
          enabled, *grid_mode* is forced to :py:attr:`.GridMode.RECURSIVE`
          (most efficient in single core)
        verbose (bool):     Print out some information
        divider (int):      Value of the divider used in
          :py:attr:`.GridMode.RECURSIVE` boundary search
        shift_zero:
        start_method (str): Python multiprocessing start method. The default
          :py:obj:`None` uses the python default that is considered safe.
          Available parameters: *'fork'*, *'spawn'*, *'forkserver'*.
          The default for linux is *'fork'*, the default for macOS and
          Windows is *'spawn'*. *'fork'* may be used for macOS to speed up
          the calculation or to solve runtime errors, however  it is
          considered unsafe.

    Returns:
        boundary:   (2,n) array: 2D acceptance
        tracked:    (2,n) array: Coordinates of tracked particles
        survived:   (2,n) array: Coordinates of surviving particles

    Example:
        >>> ring.plot_acceptance(planes, npoints, amplitudes)
        >>> plt.show()
    """
    obspt = kwargs.pop("obspt", None)
    block = kwargs.pop("block", False)
    acceptance = kwargs.pop("acceptance", None)
    if obspt is not None:
        assert np.isscalar(obspt), "Scalar value needed for obspt"
    kwargs["refpts"] = obspt
    if acceptance is None:
        boundary, survived, grid = ring.get_acceptance(planes, *args, **kwargs)
    else:
        boundary, survived, grid = acceptance
    plt.figure()
    plt.plot(*grid, ".", label="Tracked particles")
    plt.plot(*survived, ".", label="Survived particles")
    if len(planes) == 1:
        pl0 = axis_(planes[0])
        plt.plot(boundary, np.zeros(2), label="Acceptance")
        plt.title(f"1D {pl0['label']} acceptance")
        plt.xlabel(f"{pl0['label']}{pl0['unit']}")
    else:
        pl0, pl1 = axis_(planes)
        plt.plot(*boundary, label="Acceptance")
        plt.title(f"2D {pl0['label']}-{pl1['label']} acceptance")
        plt.xlabel(f"{pl0['label']}{pl0['unit']}")
        plt.xlabel(f"{pl1['label']}{pl1['unit']}")
    plt.legend()
    plt.show(block=block)
    return boundary, survived, grid


def plot_geometry(
    ring: Lattice,
    start_coordinates: tuple[float, float, float] = (0, 0, 0),
    centered: bool = False,
    ax: Axes = None,
    **kwargs,
):
    """Compute and plot the 2D ring geometry in cartesian coordinates.

    Parameters:
        ring: Lattice description
        start_coordinates: x,y,angle at starting point
        centered: it True the coordinates origin is the center of the ring
        ax: axes for the plot. If not given, new axes are created

    Keyword arguments are forwarded to the underlying
    :py:func:`~matplotlib.pyplot.plot` function

    Returns:
        geomdata: recarray containing, x, y, angle
        radius: machine radius
        ax: plot axis

    Example:
        >>> ring.plot_geometry()
    """
    if not ax:
        fig, ax = plt.subplots()
    geom, radius = ring.get_geometry(
        start_coordinates=start_coordinates, centered=centered
    )
    ax.plot(
        geom["x"],
        geom["y"],
        "o:",
        linewidth=kwargs.pop("linewidth", 0.5),
        markersize=kwargs.pop("markersize", 2),
        **kwargs,
    )
    ax.set_xlabel("x [m]")
    ax.set_ylabel("y [m]")
    ax.set_aspect("equal", "box")
    return geom, radius, ax


def plot_sigma(
    sigma,
    axis: tuple[str, str] = ("x", "xp"),
    scale: float = 1.0,
    ax: Axes = None,
    **kwargs,
):
    r"""Plot the projection of the phase space defined by a
    :math:`\Sigma`-matrix on the selected plane.

    Arguments:
        sigma:  :math:`\Sigma`-matrix
        axis:   tuple if indices defining the plane of the :math:`\Sigma`
          projection. Allowed values are: *'x'*, *'xp'*, *'y'*,
          *'yp'*, *'dp'*, *'ct'*. Default: (*'x'*, *'xp'*)
        scale:  Scaling factor for the emittance
        ax: axes for the plot. If not given, new axes are created

    Keyword arguments are forwarded to the underlying
    :py:func:`~matplotlib.pyplot.plot` function
    """
    if not ax:
        fig, ax = plt.subplots()
    ax1, ax2 = axis_(axis)
    axid = axis_(axis, key="index")
    sig22 = sigma[np.ix_(axid, axid)]
    eps = sqrt(sig22[0, 0] * sig22[1, 1] - sig22[1, 0] * sig22[0, 1])
    sigx = sqrt(sig22[0, 0])
    tr = np.array([[sigx, 0.0], [sig22[0, 1] / sigx, eps / sigx]])
    loop = 2.0 * np.pi * np.arange(0.0, 1.0, 0.001)
    normcoord = np.vstack((np.cos(loop), np.sin(loop)))
    coord = tr @ normcoord
    line = ax.plot(scale * coord[0, :], scale * coord[1, :], **kwargs)
    ax.set_title(f"{ax1['label']}-{ax2['label']} phase space")
    ax.set_xlabel(f"{ax1['label']}{ax1['unit']}")
    ax.set_ylabel(f"{ax2['label']}{ax2['unit']}")
    return line


def plot_RF_bucket_hamiltonian(
    ring,
    ct_range=None,
    dp_range=None,
    num_points=400,
    num_levels=41,
    plot_separatrix=True,
    **kwargs,
):
    r"""Plot the resulting longitudinal Hamiltonian of a ring (defining the RF
    bucket). The Hamiltonian is calculated by summing all the cavities in the
    ring. Harmonic cavities are supported by the function.

    A perfectly tuned lattice is assumed, the cavities' frequency is nominal
    and the TimeLag is set in a way ensuring ct=0 for the synchronous phase
    by using :py:func:`.set_cavity_phase`.

    Parameters:
        ring: Lattice description
        ct_range (tuple): Forced lower and upper ct values for the plot.
            Default to :math:`\pm 1.1 \times C / (2h)`
        dp_range (tuple): Forced lower and upper dp values for the plot.
            Default to twice the RF acceptance of the bucket.
        num_points (int): Number of points for 1D grid (ct/dp)
            Default to 400.
        num_levels (int): Number of levels for contour plot. Odd number of
            levels allow to center the colorbar around 0. Default to 41.
        plot_separatrix (bool): Flag to plot the separatrix contour
            (:math:`\mathcal{H}=0`).

    Returns:
        CT:   (num_points,num_points) array: ct grid
        DP:    (num_points,num_points) array: dp grid
        hamiltonian:   (num_points,num_points) array: Hamiltonian values
        along (CT,DP) grid
    """
    # Momentum compaction/phase slip factors computed up to third order
    tmp_ring = ring.disable_6d(copy=True)
    alpha = get_mcf(tmp_ring, fit_order=3, n_step=10)

    eta = np.zeros(len(alpha))
    eta[0] = alpha[0] - 1 / ring.gamma**2
    eta[1] = 3 * ring.beta**2 / 2 / ring.gamma**2 + alpha[1] - alpha[0] * eta[0]
    eta[2] = (
        -(ring.beta**2) * (5 * ring.beta**2 - 1) / (2 * ring.gamma**2)
        + alpha[2]
        - 2 * alpha[0] * alpha[1]
        + alpha[1] / ring.gamma**2
        + alpha[0] ** 2 * eta[0]
        - (3 * ring.beta**2 * alpha[0]) / (2 * ring.gamma**2)
    )

    # (ct, dp) grid calculation (defined around the main RF bucket)
    if ct_range is None:
        ct = np.linspace(
            -0.55 * ring.circumference / ring.harmonic_number,
            0.55 * ring.circumference / ring.harmonic_number,
            num=num_points,
        )
    else:
        ct = np.linspace(ct_range[0], ct_range[1], num=num_points)
    if dp_range is None:
        U0 = ring.energy_loss
        overvoltage = ring.rf_voltage / U0
        rfa = np.sqrt(
            2
            * U0
            / (np.pi * alpha[0] * ring.harmonic_number * ring.energy)
            * (np.sqrt(overvoltage**2 - 1) - np.arccos(1 / overvoltage))
        )
        dp = np.linspace(-2 * rfa, 2 * rfa, num=num_points)
    else:
        dp = np.linspace(dp_range[0], dp_range[1], num=num_points)
    CT, DP = np.meshgrid(ct, dp)

    # Hamiltonian (H=U+T) divided by harmonic number to have
    # U = U(V_rf, h, phi_s)
    # First term of the Hamiltonian
    eta_delta = eta[0] / 2 + eta[1] / 3 * DP + eta[2] / 4 * DP**2
    T = ring.beta**2 * ring.energy * eta_delta * DP**2

    hamiltonian = T
    # Iteration over all lattice cavities
    for rfcav in ring[RFCavity]:
        Voltage = rfcav.Voltage
        HarmNumber = rfcav.HarmNumber
        TimeLag = rfcav.TimeLag

        phi_s = TimeLag * 2 * np.pi * rfcav.Frequency / ring.beta / clight
        phi = (np.pi - phi_s) + CT * 2 * np.pi * rfcav.Frequency / ring.beta / clight

        # Second term of the Hamiltonian
        U = (
            Voltage
            / (2 * np.pi * HarmNumber)
            * (np.cos(phi) - np.cos(phi_s) + (phi - phi_s) * np.sin(phi_s))
        )
        # Add to total Hamiltonian
        hamiltonian += U

    fig, ax = plt.subplots(1)
    lim_range = np.max((np.abs(hamiltonian).min(), np.abs(hamiltonian).max()))
    levels = np.linspace(-lim_range, lim_range, num_levels, endpoint=True)
    co = ax.contourf(CT, DP, hamiltonian, levels, cmap="coolwarm", alpha=0.7)
    # additional contour for visibility
    ax.contour(CT, DP, hamiltonian, levels, cmap="coolwarm")
    if plot_separatrix:
        # separatrix contour
        ax.contour(CT, DP, hamiltonian, [0], colors="black")
        plt.plot([], [], color="black", label="Separatrix")
        ax.legend()
    cb = fig.colorbar(co)
    cb.set_label(r"$\mathcal{H}(ct,\delta)$ [a.u.]", fontsize=18)

    ax.set_xlabel(r"ct [m]")
    ax.set_ylabel(r"$\delta$")

    phi_s = (
        ring.get_rf_timelag()
        * 2
        * np.pi
        * ring.get_revolution_frequency()
        * ring.harmonic_number
        / (ring.beta * clight)
    )

    def ct_to_phi(ct):
        return (
            np.pi
            - phi_s
            + ct
            / (
                2
                * np.pi
                * ring.get_revolution_frequency()
                * ring.harmonic_number
                / clight
            )
        )

    def phi_to_ct(phase):
        return (
            np.pi
            - phi_s
            - phase
            * (
                2
                * np.pi
                * ring.get_revolution_frequency()
                * ring.harmonic_number
                / clight
            )
        )

    ax2 = ax.secondary_xaxis("top", functions=(phi_to_ct, ct_to_phi))
    ax2.set_xlabel(r"$\phi$ [rad]")

    plt.title(r"$\phi_{RF}$ " + rf"= $\pi -$ {phi_s:.2f}", fontsize=18)

    return CT, DP, hamiltonian


Lattice.plot_acceptance = plot_acceptance
Lattice.plot_geometry = plot_geometry
Lattice.plot_RF_bucket_hamiltonian = plot_RF_bucket_hamiltonian



"""AT plotting functions"""

try:
    # noinspection PyPackageRequirements
    import matplotlib
except (ImportError, RuntimeError) as exc:
    print("matplotlib is unavailable => plotting is disabled.")
    print("To enable plotting functions, run \"pip install matplotlib\".")
else:
    from .synopt import *
    from .generic import *
    from .specific import *
    from .standalone import *
    from .resonances import *
    from .response_matrix import *



"""Helper functions for axis and plane descriptions"""

from __future__ import annotations

# Necessary for type aliases in python <= 3.8 :
from typing import Tuple
from typing import Union

AxisCode = Union[str, int, slice, None, type(Ellipsis)]
AxisDef = Union[AxisCode, Tuple[AxisCode, AxisCode]]

_axis_def = {
    "x": {"index": 0, "label": "x", "unit": " [m]"},
    "px": {"index": 1, "label": r"$p_x$", "unit": " [rad]"},
    "y": {"index": 2, "label": "y", "unit": " [m]"},
    "py": {"index": 3, "label": r"$p_y$", "unit": " [rad]"},
    "dp": {"index": 4, "label": r"$\delta$", "unit": ""},
    "ct": {"index": 5, "label": r"$\beta c \tau$", "unit": " [m]"},
}
for xk, xv in list(_axis_def.items()):
    xv["code"] = xk
    _axis_def[xv["index"]] = xv
    _axis_def[xk.upper()] = xv
_axis_def["delta"] = _axis_def["dp"]
_axis_def["xp"] = _axis_def["px"]  # For backward compatibility
_axis_def["yp"] = _axis_def["py"]  # For backward compatibility
_axis_def["s"] = _axis_def["ct"]
_axis_def["S"] = _axis_def["ct"]
_axis_def[None] = {"index": None, "label": "", "unit": "", "code": ":"}
_axis_def[Ellipsis] = {"index": Ellipsis, "label": "", "unit": "", "code": "..."}

_plane_def = {
    "x": {"index": 0, "label": "x", "unit": " [m]"},
    "y": {"index": 1, "label": "y", "unit": " [m]"},
    "z": {"index": 2, "label": "z", "unit": ""},
}
for xk, xv in list(_plane_def.items()):
    xv["code"] = xk
    _plane_def[xv["index"]] = xv
    _plane_def[xk.upper()] = xv
_plane_def["h"] = _plane_def["x"]
_plane_def["v"] = _plane_def["y"]
_plane_def["H"] = _plane_def["x"]
_plane_def["V"] = _plane_def["y"]
_plane_def[None] = {"index": None, "label": "", "unit": "", "code": ":"}
_plane_def[Ellipsis] = {"index": Ellipsis, "label": "", "unit": "", "code": "..."}


def _descr(dd: dict, *args: AxisDef, key: str | None = None):
    for arg in args:
        if isinstance(arg, tuple):
            for a in arg:
                yield from _descr(dd, a, key=key)
        else:
            if isinstance(arg, slice):
                descr = {"index": arg, "code": arg, "label": "", "unit": ""}
            else:
                descr = dd[arg]
            if key is None:
                yield descr
            else:
                yield descr[key]


def axis_(*axis: AxisDef, key: str | None = None):
    r"""Return axis descriptions

    Parameters:
        axis:           axis code or tuple of axis codes selecting axes in the
          standard AT coordinate system. codes for the 6 axes are:

          0, 'x', 'X'

          1, 'px', PX'

          2, 'y', 'Y'

          3, 'py', 'PY'

          4, 'dp', 'DP', 'delta'

          5, 'ct', 'CT', 's', 'S'

          :py:obj:`None` and :py:obj:`Ellipsis` select all axes

        key:            key in the coordinate descriptor dictionary,
          selecting the desired information. One of :

          'index'
            index in the standard AT coordinate vector
          'code'
            string representation
          'label'
            label for plot annotation
          'unit'
            coordinate unit
          :py:obj:`None`
            entire description dictionary

    Returns:
        descr : value or tuple[values]

    Examples:

        >>> axis_("x", "dp", key="index")
        (0, 4)

        returns the indices in the standard coordinate vector

        >>> dplabel = axis_("dp", key="label")
        >>> print(dplabel)
        $\delta$

        returns the coordinate label for plot annotation

        >>> axis_(0, "dp")
        ({'plane': 0, 'label': 'x', 'unit': ' [m]', 'code': 'x'},
         {'plane': 4, 'label': '$\\delta$', 'unit': '', 'code': 'dp'})

        returns the entire description directories

    """
    ret = tuple(_descr(_axis_def, *axis, key=key))
    if len(ret) > 1:
        return ret
    else:
        return ret[0]


def plane_(*plane: AxisDef, key: str | None = None):
    r"""Return plane descriptions

    Parameters:
        plane:          plane code or tuple of plane codes selecting planes.
          codes for the 3 planes are:

          0, 'x', 'X', 'h', 'H' for horizontal plane,

          1, 'y', 'Y', 'v', 'V' for vertical plane,

          2, 'z', 'Z'   for the longitudinal plane

          :py:obj:`None`, slice(None) and :py:obj:`Ellipsis` selects all planes
        key:            key in the plane descriptor dictionary,
          selecting the desired information. One of :

          'plane'
            plane in the optics data
          'code'
            string representation
          'label'
            label for plot annotation
          'unit'
            coordinate unit
          :py:obj:`None`
            entire description dictionary

    Returns:
        descr : value or tuple[values]

    Examples:

        >>> plane_("v", key="index")
        1

        returns the indices in the standard coordinate vector

        >>> plane_("x", "y")
        ({'plane': 0, 'label': 'x', 'unit': ' [m]', 'code': 'h'},
         {'plane': 1, 'label': 'y', 'unit': ' [m]', 'code': 'v'})

        returns the entire description directories

    """
    ret = tuple(_descr(_plane_def, *plane, key=key))
    if len(ret) > 1:
        return ret
    else:
        return ret[0]



# noinspection PyUnresolvedReferences
r"""Using `MAD-X`_ files with PyAT
==================================

PyAT can read lattice descriptions in Mad-X format (.seq files), and can export
lattices in MAD-X format.

However, because of intrinsic differences between PyAT and MAD-X, some
incompatibilities must be taken into account.

1. Translation losses
---------------------

6-D motion
^^^^^^^^^^^^^^
While AT allows setting 6-D motion and synchrotron radiation on individual elements,
MAD has a global setting for the whole lattice. When reading a MAD-X lattice without
radiation, 6-D motion in the resulting AT lattice is turned off, including in RF
cavities. If the MAD-X lattice has radiation on, 6-D motion is activated on the AT
lattice according to default settings (RF cavities active, radiation in dipoles,
quadrupoles and wigglers).

Combined function magnets
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

MAD has little support for combined function magnets. When exporting a lattice in MAD
format, the main field component for each magnet class in kept but other components
disappear, except in the few cases handled by MAD (quadrupole and sextupole
components in ``SBEND`` or ``RBEND``, skew quad component in ``QUADRUPOLE``, see the
MAD user's reference manual for more).

Multipoles
^^^^^^^^^^^^^^

MAD has no thick multipoles. Multipoles in MAD format (``MULTIPOLE`` element) are
interpreted as :py:class:`.ThinMultipole` elements. In the other direction, an AT
:py:class:`.Multipole` is converted to a thin ``MULTIPOLE`` surrounded by two
drift spaces.

MAD elements absent from AT
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
Many MAD elements have no equivalent in AT. They are replaced by
:py:class:`.Marker` or :py:class:`.Drift` elements, depending on their length.

Incompatible attributes
^^^^^^^^^^^^^^^^^^^^^^^^^^^
Some AT element attributes have no MAD equivalent, and vice versa.

When exporting to a MAD-X file:

- `NumIntSteps`, `MaxOrder` are discarded,
- `FringeBendEntrance`, `FringeBendExit`, `FringeQuadEntrance`, `FringeQuadExit` are
  discarded,
- `R1`, `R2`, `T1`, `T2` are discarded,
- `RApertures`, `EApertures` are discarded.

When reading a MAD-X file:

- `TILT` is interpreted and converted to `R1` and `R2` attributes,
- `KTAP` is interpreted and converted to `FieldScaling`.

.. _using-mad-x-files:

2. Reading MAD-X files
----------------------

Short way
^^^^^^^^^
The :py:func:`load_madx` function created a :py:class:`.Lattice` from a ``LINE`` or
``SEQUENCE`` in a MAD-X file.

>>> ring = at.load_madx("lattice.seq", use="PS")

>>> print(ring)
Lattice(<1295 elements>, name='PS', energy=1000000000.0, particle=Particle('positron'),
periodicity=1, harmonic_number=0, beam_current=0.0, nbunch=1, use='PS')

Detailed way
^^^^^^^^^^^^
:py:class:`MadxParser` creates an empty database which can be populated with the
elements of a MAD-X file.

>>> parser = at.MadxParser()

The :py:meth:`MadxParser.parse_files` method reads one or several MAD-X files and
populates the parser

>>> parser.parse_files("lattice.seq", use="PS")

The parser can be examined and modified using the standard python syntax:

>>> parser["pr_bht91_f"]
sbend(name=PR.BHT91.F, l=2.1975925, angle='angle.prbhf', k1='k1prbhf+k1prpfwf-k1prf8l')

>>> parser["angle.prbhf"]
0.03135884818

>>> parser["angle.prbhf"] = 0.032

All MAD parameters can be interactively modified and their last value will be taken
into account when generating a PyAT lattice.

The :py:meth:`MadxParser.lattice` method creates a :py:class:`.Lattice` from a ``LINE``
or ``SEQUENCE`` of the parser:

>>> ring = parser.lattice(use="ps")

>>> print(ring)
Lattice(<1295 elements>, name='PS', energy=1000000000.0, particle=Particle('positron'),
periodicity=1, harmonic_number=0, beam_current=0.0, nbunch=1, use='PS')

3. Exporting to MAD-X files
---------------------------
Exporting a PyAT lattice to a MAD-X files produces a single MAD ``SEQUENCE`` of
``LINE``.

See :py:func:`save_madx` for usage.

4. Functions and classes
------------------------

.. _mad-x: https://mad.web.cern.ch/mad/webguide/manual.html
"""

from __future__ import annotations

__all__ = ["MadParameter", "MadxParser", "load_madx", "save_madx"]

import functools
import warnings

# functions known by MAD-X
from math import pi, e, sqrt, exp, log, log10, sin, cos, tan  # noqa: F401
from math import asin, acos, atan, sinh, cosh, tanh, erf, erfc  # noqa: F401
from itertools import chain
from collections.abc import Sequence, Generator, Iterable
import re

import numpy as np

# constants known by MAD-X
from scipy.constants import c as clight, hbar as _hb, e as qelect
from scipy.constants import physical_constants as _cst

from .allfiles import register_format
from .utils import split_ignoring_parentheses, protect, restore
from .file_input import AnyDescr, ElementDescr, SequenceDescr, BaseParser
from .file_input import LowerCaseParser, UnorderedParser
from .file_input import set_argparser, ignore_names
from .file_output import Exporter
from ..lattice import Lattice, Particle, elements as elt, tilt_elem

_separator = re.compile(r"(?<=[\w.)])\s+(?=[\w.(])")

# Constants known by MAD-X
true = True
false = False
twopi = 2 * pi
degrad = 180.0 / pi
raddeg = pi / 180.0
emass = 1.0e-03 * _cst["electron mass energy equivalent in MeV"][0]  # [GeV]
pmass = 1.0e-03 * _cst["proton mass energy equivalent in MeV"][0]  # [GeV]
nmass = 1.0e-03 * _cst["neutron mass energy equivalent in MeV"][0]  # [GeV]
umass = 1.0e-03 * _cst["atomic mass constant energy equivalent in MeV"][0]  # [GeV]
mumass = 1.0e-03 * _cst["muon mass energy equivalent in MeV"][0]  # [GeV]
hbar = _hb / qelect * 1.0e-09  # [GeV.s]
erad = _cst["classical electron radius"][0]  # [m]
prad = erad * emass / pmass  # [m]


class MadParameter(str):
    """MAD parameter

    A MAD parameter is an expression which can be evaluated n the context
    of a MAD parser
    """

    def __new__(cls, parser, expr):
        return super().__new__(cls, expr)

    # noinspection PyUnusedLocal
    def __init__(self, parser: _MadParser, expr: str):
        """Args:
            parser: MadParser instance defining the context for evaluation
            expr:   expression to be evaluated

        The expression may contain MAD parameter names, arithmetic operators and
        mathematical functions known by MAD
        """
        self.parser = parser

    def __float__(self):
        return float(self.parser.evaluate(self))

    def __int__(self):
        return int(self.parser.evaluate(self))

    def __add__(self, other):
        return float(self) + float(other)

    def __radd__(self, other):
        return float(other) + float(self)

    def __mul__(self, other):
        return float(self) * float(other)

    def __rmul__(self, other):
        return float(other) * float(self)

    def __sub__(self, other):
        return float(self) - float(other)

    def __rsub__(self, other):
        return float(other) - float(self)

    def __truediv__(self, other):
        return float(self) / float(other)

    def __rtruediv__(self, other):
        return float(other) / float(self)

    def __pow__(self, other):
        return pow(float(self), other)

    def __rpow__(self, other):
        return pow(float(other), float(self))

    def __neg__(self):
        return -float(self)

    def __pos__(self):
        return +float(self)

    def evaluate(self):
        return self.parser.evaluate(self)


def sinc(x: float) -> float:
    """Cardinal sine function known by MAD-X"""
    if abs(x) < 1e-10:
        return x
    else:
        return sin(x) / x


# -------------------
#  Utility functions
# -------------------


def mad_element(func):
    """Decorator for AT elements"""

    @functools.wraps(func)
    def wrapper(self, *args, tilt=0.0, ktap=0.0, **kwargs):
        elems = func(self, *args, **kwargs)
        for el in elems:
            tilt = float(tilt)  # MadParameter conversion
            if tilt != 0.0:
                tilt_elem(el, tilt)
            ktap = float(ktap)  # MadParameter conversion
            if ktap != 0.0:
                el.Scaling = 1.0 + ktap
            el.origin = self.origin
        return elems

    return wrapper


def poly_to_mad(x: Iterable[float], factor: float = 1.0) -> Generator[float]:
    """Convert polynomials from AT to MAD"""
    f = 1.0
    for n, vx in enumerate(x):
        yield factor * float(vx * f)
        f *= n + 1


def poly_from_mad(x: Iterable[float], factor: float = 1.0) -> Generator[float]:
    """Convert polynomials from MAD to AT"""
    f = 1.0
    for n, vx in enumerate(x):
        yield factor * float(vx / f)
        f *= n + 1


def p_to_at(a: float | Sequence[float]) -> np.ndarray:
    """Convert polynomials from MADX to AT"""
    if not isinstance(a, Sequence):
        # In case of a single element, we have a scalar instead of a tuple
        a = (a,)
    return np.fromiter(poly_from_mad(a), dtype=float)


def p_dict(keys: Iterable[str], a: Iterable[float]) -> dict[str, float]:
    """return K1, K2... from an AT Polynom"""
    return {k: v for k, v in zip(keys, poly_to_mad(a)) if k and (v != 0.0)}


def p_list(a: Iterable[float], factor: float = 1.0):
    """Return a Polynom list"""
    return list(poly_to_mad(a, factor=factor))


# noinspection PyUnusedLocal
def _keyparser(parser, argcount, argstr):
    """Return the pair key, value for the given 'key' argument"""
    return argstr, parser.evaluate(argstr)


# ------------------------------
#  Base class for MAD-X elements
# ------------------------------


class _MadElement(ElementDescr):
    """Description of MADX elements"""

    str_attr = {"apertype", "refer", "refpos", "sequence", "from"}

    def __init__(self, *args, at=0.0, **kwargs):
        self.at = at
        # Cannot use "from" as argument or attribute
        setattr(self, "from", kwargs.pop("from", None))
        # kwargs.pop("copy", False)
        super().__init__(*args, **kwargs)

    def limits(self, parser: MadxParser, offset: float, refer: float):
        half_length = 0.5 * self.length
        offset = offset + refer * half_length + self.at
        frm = getattr(self, "from")
        if frm is not None:
            offset += parser[frm].at
        return np.array([-half_length, half_length]) + offset

    def meval(self, params: dict):
        """Evaluation of superfluous parameters"""

        def mpeval(v):
            if isinstance(v, MadParameter):
                return v.evaluate()
            elif isinstance(v, str):
                return v
            elif isinstance(v, Sequence):
                return tuple(mpeval(a) for a in v)
            else:
                return v

        # return {k: mpeval(v) for k, v in params.items()}
        return {}


# ------------------------------
#  MAD-X element classes
# ------------------------------


# noinspection PyPep8Naming
class drift(_MadElement):
    @mad_element
    def to_at(self, l=0.0, **params):  # noqa: E741
        return [elt.Drift(self.name, l, **self.meval(params))]


# noinspection PyPep8Naming
class marker(_MadElement):
    at2mad = {}

    @mad_element
    def to_at(self, **params):
        return [elt.Marker(self.name, **self.meval(params))]


# noinspection PyPep8Naming
class quadrupole(_MadElement):
    @mad_element
    def to_at(self, l, k1=0.0, k1s=0.0, **params):  # noqa: E741
        atparams = {}
        k1s = float(k1s)  # MadParameter conversion  # MadParameter conversion
        if k1s != 0.0:
            atparams["PolynomA"] = [0.0, k1s]
        return [elt.Quadrupole(self.name, l, k1, **atparams, **self.meval(params))]

    @classmethod
    def from_at(cls, kwargs):
        el = super().from_at(kwargs)
        el.update(p_dict(["K0", "K1"], kwargs.pop("PolynomB", ())))
        el.update(p_dict(["K0S", "K1S"], kwargs.pop("PolynomA", ())))
        return el


# noinspection PyPep8Naming
class sextupole(_MadElement):
    @mad_element
    def to_at(self, l, k2=0.0, k2s=0.0, **params):  # noqa: E741
        atparams = {}
        k2s = float(k2s)  # MadParameter conversion
        if k2s != 0.0:
            atparams["PolynomA"] = [0.0, 0.0, k2s / 2.0]
        return [elt.Sextupole(self.name, l, k2 / 2.0, **atparams, **self.meval(params))]

    @classmethod
    def from_at(cls, kwargs):
        el = super().from_at(kwargs)
        el.update(p_dict(["K0", "K1", "K2"], kwargs.pop("PolynomB", ())))
        el.update(p_dict(["K0S", "K1S", "K2S"], kwargs.pop("PolynomA", ())))
        return el


# noinspection PyPep8Naming
class octupole(_MadElement):
    @mad_element
    def to_at(self, l, k3=0.0, k3s=0.0, **params):  # noqa: E741
        poly_b = [0.0, 0.0, 0.0, k3 / 6.0]
        poly_a = [0.0, 0.0, 0.0, k3s / 6.0]
        return [elt.Octupole(self.name, l, poly_a, poly_b, **self.meval(params))]

    @classmethod
    def from_at(cls, kwargs):
        el = super().from_at(kwargs)
        el.update(p_dict(["K0", "K1", "K2", "K3"], kwargs.pop("PolynomB", ())))
        el.update(p_dict(["K0S", "K1S", "K2S", "K3S"], kwargs.pop("PolynomA", ())))
        return el


# noinspection PyPep8Naming
class multipole(_MadElement):
    at2mad = {}

    @mad_element
    def to_at(self, knl=(), ksl=(), **params):
        params.pop("l", None)
        return [
            elt.ThinMultipole(
                self.name, p_to_at(ksl), p_to_at(knl), **self.meval(params)
            )
        ]

    @classmethod
    def from_at(cls, kwargs, factor=1.0):
        el = super().from_at(kwargs)
        el["KNL"] = p_list(kwargs.pop("PolynomB", ()), factor=factor)
        el["KSL"] = p_list(kwargs.pop("PolynomA", ()), factor=factor)
        return el


# noinspection PyPep8Naming
class sbend(_MadElement):
    at2mad = {
        "Length": "L",
        "BendingAngle": "ANGLE",
        "EntranceAngle": "E1",
        "ExitAngle": "E2",
    }

    @mad_element
    def to_at(
        self,
        l,  # noqa: E741
        angle,
        e1=0.0,
        e2=0.0,
        k1=0.0,
        k2=0.0,
        k1s=0.0,
        hgap=None,
        fint=0.0,
        **params,
    ):
        atparams = {}
        if hgap is not None:
            fintx = params.pop("fintx", fint)
            atparams.update(FullGap=2.0 * hgap, FringeInt1=fint, FringeInt2=fintx)
        k2 = float(k2)  # MadParameter conversion
        if k2 != 0.0:
            atparams["PolynomB"] = [0.0, k1, k2 / 2.0]
        k1s = float(k1s)  # MadParameter conversion
        if k1s != 0.0:
            atparams["PolynomA"] = [0.0, k1s]

        return [
            elt.Dipole(
                self.name,
                l,
                angle,
                k1,
                EntranceAngle=e1,
                ExitAngle=e2,
                **atparams,
                **self.meval(params),
            )
        ]

    @classmethod
    def from_at(cls, kwargs):
        el = super().from_at(kwargs)
        el.update(p_dict(["K0", "K1", "K2"], kwargs.pop("PolynomB", ())))
        el.update(p_dict(["K0S", "K1S", "K2S"], kwargs.pop("PolynomA", ())))
        return el


# noinspection PyPep8Naming
class rbend(sbend):
    @mad_element
    def to_at(self, l, angle, e1=0.0, e2=0.0, **params):  # noqa: E741
        hangle = abs(0.5 * angle)
        arclength = l / sinc(hangle)
        return super().to_at(arclength, angle, e1=hangle + e1, e2=hangle + e2, **params)

    @property
    def length(self):
        """Element length"""
        hangle = 0.5 * self["angle"]
        return self["l"] / sinc(hangle)


# noinspection PyPep8Naming
class kicker(_MadElement):
    @mad_element
    def to_at(self, l=0.0, hkick=0.0, vkick=0.0, **params):  # noqa: E741
        kicks = np.array([hkick, vkick], dtype=float)
        return [elt.Corrector(self.name, l, kicks, **self.meval(params))]

    @classmethod
    def from_at(cls, kwargs):
        el = super().from_at(kwargs)
        kicks = kwargs.pop("KickAngle", (0.0, 0.0))
        el["HKICK"] = kicks[0]
        el["VKICK"] = kicks[1]
        return el


# noinspection PyPep8Naming
class hkicker(kicker):
    @mad_element
    def to_at(self, l=0.0, kick=0.0, **params):  # noqa: E741
        return super().to_at(l=l, hkick=kick, **params)


# noinspection PyPep8Naming
class vkicker(kicker):
    @mad_element
    def to_at(self, l=0.0, kick=0.0, **params):  # noqa: E741
        return super().to_at(l=l, vkick=kick, **params)


# noinspection PyPep8Naming
class rfcavity(_MadElement):
    @mad_element
    def to_at(
        self,
        l=0.0,  # noqa: E741
        volt=0.0,
        freq=np.nan,
        lag=0.0,
        harmon=0,
        **params,
    ):
        cavity = elt.RFCavity(
            self.name,
            l,
            1.0e6 * volt,
            1.0e6 * freq,
            harmon,
            0.0,
            PassMethod="IdentityPass" if float(l) == 0.0 else "DriftPass",
            **self.meval(params),
        )
        return [cavity]

    @classmethod
    def from_at(cls, kwargs):
        el = super().from_at(kwargs)
        el["VOLT"] = 1.0e-6 * kwargs.pop("Voltage")
        el["FREQ"] = 1.0e-6 * kwargs.pop("Frequency")
        return el


# noinspection PyPep8Naming
class monitor(_MadElement):
    @mad_element
    def to_at(self, l=0.0, **params):  # noqa: E741
        hl = 0.5 * l  # MadParameter conversion
        if hl == 0.0:
            return [elt.Monitor(self.name, **self.meval(params))]
        else:
            drname = self.name + ".1"
            return [
                elt.Drift(drname, hl),
                elt.Monitor(self.name, **self.meval(params)),
                elt.Drift(drname, hl),
            ]


# noinspection PyPep8Naming
class hmonitor(monitor):
    pass


# noinspection PyPep8Naming
class vmonitor(monitor):
    pass


# noinspection PyPep8Naming
class instrument(monitor):
    pass


ignore_names(
    globals(),
    _MadElement,
    ["solenoid", "rfmultipole", "crabcavity", "elseparator", "collimator", "tkicker"],
)


@set_argparser(_keyparser)
def value(**kwargs):
    """VALUE command"""
    kwargs.pop("copy", False)
    for key, v in kwargs.items():
        print(f"{key}: {v}")


class _Line(SequenceDescr):
    """Descriptor for the MADX LINE"""

    def __add__(self, other):
        return type(self)(chain(self, other))

    def __mul__(self, other):
        def repeat(n):
            for _i in range(n):
                yield from self

        return type(self)(repeat(other))

    def __rmul__(self, other):
        return self.__mul__(other)

    def expand(self, parser: BaseParser) -> Generator[elt.Element, None, None]:
        if self.inverse:
            for elem in reversed(self):
                if isinstance(elem, AnyDescr):  # Element or List
                    yield from (-elem).expand(parser)
                elif isinstance(elem, Sequence):  # Other sequence (tuple)
                    for el in reversed(elem):
                        yield from (-el).expand(parser)
        else:
            for elem in self:
                if isinstance(elem, AnyDescr):  # Element or List
                    yield from elem.expand(parser)
                elif isinstance(elem, Sequence):  # Other sequence (tuple, list)
                    for el in elem:
                        yield from el.expand(parser)


# noinspection PyPep8Naming
class _Sequence(SequenceDescr):
    """Descriptor for the MADX SEQUENCE"""

    _offset = {"entry": 1.0, "centre": 0.0, "exit": -1.0}

    def __init__(
        self,
        *args,
        l: float = 0,  # noqa: E741
        refer: str = "centre",
        refpos: str | None = None,
        at: float = 0.0,
        valid: int = 1,
        **kwargs,
    ):
        self.l = l  # noqa: E741
        try:
            self.refer = self._offset[refer]
        except KeyError as exc:
            raise ValueError(f"REFER must be in {set(self._offset.keys())}") from exc
        self.refpos = refpos
        self.at = at
        # Cannot use "from" as argument or attribute name:
        setattr(self, "from", kwargs.pop("from", None))
        self.valid = bool(valid)
        super().__init__(*args, **kwargs)

    def __call__(self, *args, copy: bool = True, **kwargs):
        # Never make a copy
        super().__call__(*args, copy=False, **kwargs)
        return self if copy else None

    def reference(self, parser, refer, refpos):
        if refpos is None:
            orig = 0.5 * (refer - 1.0) * self.length
        else:
            orig = None
            for fname, *anames in self:
                if fname == refpos:
                    # noinspection PyProtectedMember
                    elem = parser._raw_command(
                        None, fname, *anames, no_global=True, copy=True
                    )
                    orig = -elem.at
                    break
            if orig is None:
                raise NameError(
                    f"REFPOS {refpos!r} is not in the sequence {self.name!r}"
                )
        frm = getattr(self, "from")
        if frm is not None:
            orig += parser[frm].at
        return orig

    def flatten(self, parser, offset: float = 0.0, refer: float = 1.0, refpos=None):
        offset = offset + self.reference(parser, refer, refpos) + self.at
        for fname, *anames in self:
            try:
                # noinspection PyProtectedMember
                elem = parser._raw_command(
                    None, fname, *anames, no_global=True, copy=True
                )
            except Exception as exc:
                mess = (f"In sequence {self.name!r}: \"{fname}, {', '.join(anames)}\"",)
                exc.args += mess
                raise
            if isinstance(elem, _Sequence):
                yield from elem.flatten(parser, offset, self.refer, elem.refpos)
            elif isinstance(elem, _MadElement):
                yield elem.limits(parser, offset, self.refer), elem

    def expand(self, parser: MadxParser) -> Generator[elt.Element, None, None]:
        def insert_drift(dl, el):
            nonlocal drcounter
            if dl > 1.0e-5:
                yield from drift(name=f"drift_{drcounter}", l=dl).expand(parser)
                drcounter += 1
            elif dl < -1.0e-5:
                elemtype = type(el).__name__.upper()
                raise ValueError(f"{elemtype}({el.name!r}) is overlapping by {-dl} m")

        drcounter = 0
        end = 0.0
        elem = self
        self.at = 0.0
        for (entry, ext), elem in self.flatten(parser):
            yield from insert_drift(entry - end, elem)
            end = ext
            yield from elem.expand(parser)

        yield from insert_drift(self.length - end, elem)  # Final drift


class _BeamDescr(ElementDescr):
    """Descriptor for the MAD-X BEAM object"""

    @staticmethod
    def to_at(name, *args, **params):
        return []

    def expand(self, parser: MadxParser) -> dict:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            atparticle = Particle(
                self["particle"],
                rest_energy=self.get("mass", emass),
                charge=self.get("charge", 1),
            )
        mass = 1.0e-09 * atparticle.rest_energy  # [GeV]
        charge = atparticle.charge

        # Respect the precedence of MAD-X
        if "energy" in self:
            energy = float(self["energy"])  # force evaluation
            gamma = energy / mass
            # betagamma = sqrt(gamma * gamma - 1.0)
        elif "pc" in self:
            pc = float(self["pc"])  # force evaluation
            betagamma = pc / mass
            gamma = sqrt(betagamma * betagamma + 1.0)
        elif "gamma" in self:
            gamma = float(self["gamma"])  # force evaluation
            # betagamma = sqrt(gamma * gamma - 1.0)
        elif "beta" in self:
            beta = float(self["beta"])  # force evaluation
            gamma = 1.0 / sqrt(1.0 - beta * beta)
            # betagamma = beta * gamma
        elif "brho" in self:
            brho = float(self["brho"])  # force evaluation
            betagamma = 1.0e-9 * brho * clight * abs(charge) / mass
            gamma = sqrt(betagamma * betagamma + 1.0)
        else:
            gamma = 1.0 / mass

        # convert from array scalar to float
        energy = float(gamma * mass)  # [GeV]
        # pc = betagamma * mass  # [GeV]
        # beta = betagamma / gamma
        # brho = 1.0e09 * pc / abs(charge) / clight

        return {
            "particle": atparticle,
            "energy": 1.0e09 * energy,  # [eV]
            "beam_current": float(self["bcurrent"]),  # force deferred evaluation
            "nbunch": self["kbunch"],
            "periodicity": 1,
            "radiate": self["radiate"],
        }


class _Call:
    """Implement the CALL Mad command"""

    @staticmethod
    def argparser(parser, argcount, argstr):
        return parser._argparser(
            argcount, argstr, pos_args=("file",), str_attr=("file",)
        )

    def __init__(self, parser: _MadParser):
        self.parser = parser

    def __call__(self, file=None, name=None, copy=False):
        self.parser.parse_files(file, final=False)


class _Beam:
    """Implement the BEAM Mad command"""

    default_beam = {
        "particle": "positron",
        "energy": 1.0,  # GeV
        "bcurrent": 0.0,
        "kbunch": 1,
        "radiate": False,
    }

    @staticmethod
    def argparser(parser, argcount, argstr):
        return parser._argparser(
            argcount,
            argstr,
            str_attr=("particle", "sequence"),
            bool_attr=("radiate", "bunched"),
        )

    def __init__(self, parser: _MadParser):
        self.parser = parser

    def __call__(self, sequence=None, **kwargs):
        """create a :py:class:`_BeamDescr` object and store it as 'beam%sequence'"""
        name = "beam%" if sequence is None else f"beam%{sequence}"
        beamobj = self.parser.get(name, None)
        if beamobj is None:
            beamobj = _BeamDescr(self.default_beam)
            self.parser[name] = beamobj

        for k, v in kwargs.items():
            beamobj[k] = v


class _MadParser(LowerCaseParser, UnorderedParser):
    """Common class for both MAD8 anf MAD-X parsers"""

    _delimiter = ";"
    _linecomment = ("!", "//")
    _endfile = "return"
    _undef_key = "none"

    _str_arguments = {"file", "refer", "refpos", "sequence", "from"}

    def __init__(self, env: dict, **kwargs):
        """Common behaviour for MAD-X and MAD8

        Args:
            env: global namespace used for evaluating commands
            verbose:    If :py:obj:`True`, print details on the processing
            strict: If :py:obj:`False`, assign 0 to undefined variables
            **kwargs: Initial variable definitions
        """
        super().__init__(
            env,
            call=_Call(self),
            beam=_Beam(self),
            sequence=_Sequence,
            centre="centre",
            entry="entry",
            exit="exit",
            **kwargs,
        )
        self.current_sequence = None
        self["beam"]()

    def clear(self):
        super().clear()
        self.current_sequence = None
        self["beam"]()

    def _assign_deferred(self, value: str):
        """Deferred assignment"""
        if value[0] == "(" and value[-1] == ")":
            # Array variable: convert to tuple
            value, matches = protect(value[1:-1], fence=(r"\(", r"\)"))
            return tuple(
                MadParameter(self, v) for v in restore(matches, *value.split(","))
            )
        else:
            # Scalar variable
            return MadParameter(self, value)

    def _argparser(self, argcount, argstr: str, **kwargs):
        key, *value = split_ignoring_parentheses(
            argstr, delimiter=":=", fence=('"', '"'), maxsplit=1
        )
        if value:
            return key, self._assign_deferred(value[0])
        else:
            return super()._argparser(argcount, argstr, **kwargs)

    def _assign(self, label: str, key: str, val: str):
        # Special treatment of "line=(...)" assignments
        if key == "line":
            val = val.replace(")", ",)")  # For tuples with a single item
            return label, _Line(self.evaluate(val), name=label)
        else:
            return super()._assign(label, key, val)

    def _command(self, label: str | None, cmdname: str, *argnames: str, **kwargs):
        # Special treatment of SEQUENCE definitions
        res = None
        if cmdname == "endsequence":
            self.current_sequence = None
            return None
        if self.current_sequence is None:
            try:
                res = super()._command(label, cmdname, *argnames, **kwargs)
            except (KeyError, NameError) as exc:
                if cmdname == "sequence":
                    # if sequence creation failed, create a dummy sequence anyway
                    res = super()._command(label, cmdname, "valid=0", **kwargs)
                    # But store the command for later update
                    self._fallback(exc, None, label, "valid=1", *argnames)
                else:
                    raise
            finally:
                if isinstance(res, _Sequence):
                    self.current_sequence = res
        else:
            if label is None:
                self.current_sequence.append((cmdname, *argnames))
            else:
                try:
                    res = super()._command(label, cmdname, *argnames, **kwargs)
                finally:
                    self.current_sequence.append((label, *argnames))
        return res

    def _decode(self, label: str, cmdname: str, *argnames: str) -> None:
        left, *right = cmdname.split(":=")
        if right:
            self[left] = self._assign_deferred(right[0])
        else:
            super()._decode(label, cmdname, *argnames)

    def _format_statement(self, line: str) -> str:
        line = _separator.sub(",", line)  # Replace space separators with commas
        # turn curly braces into parentheses. Must be done before splitting arguments
        line = line.replace("{", "(").replace("}", ")")
        return super()._format_statement(line)

    def _get_beam(self, key: str):
        """Get the beam object for a given sequence"""
        try:
            beam = self[f"beam%{self._gen_key(key)}"]
        except KeyError:
            beam = self["beam%"]
        return beam

    def _generator(self, params: dict) -> Iterable[elt.Element]:
        def beta() -> float:
            rest_energy = params["particle"].rest_energy
            if rest_energy == 0.0:
                return 1.0
            else:
                gamma = float(params["energy"] / rest_energy)
                return sqrt(1.0 - 1.0 / gamma / gamma)

        # generate the Lattice attributes from the BEAM object
        beam = self._get_beam(params["use"]).expand(self)
        for key, val in beam.items():
            params.setdefault(key, val)

        cavities = []
        cell_length = 0

        for elem in super()._generator(params):
            if isinstance(elem, elt.RFCavity):
                cavities.append(elem)
            cell_length += getattr(elem, "Length", 0.0)
            yield elem

        rev = beta() * clight / cell_length

        # Set the frequency of cavities in which it is not specified
        for cav in cavities:
            if np.isnan(cav.Frequency):
                cav.Frequency = rev * cav.HarmNumber
            elif cav.HarmNumber == 0:
                cav.HarmNumber = round(cav.Frequency / rev)

    def lattice(self, use: str = "ring", **kwargs):
        """Create a lattice from the selected sequence

        Parameters:
            use:                Name of the MAD sequence or line containing the desired
              lattice. Default: ``ring``

        Keyword Args:
            name (str):         Name of the lattice. Default: MAD sequence name.
            particle(Particle): Circulating particle. Default: from MAD
            energy (float):     Energy of the lattice [eV]. Default: from MAD
            periodicity(int):   Number of periods. Default: 1
            *:                  All other keywords will be set as Lattice attributes
        """
        part = kwargs.get("particle", None)
        if isinstance(part, str):
            kwargs["particle"] = Particle(part)

        lat = super().lattice(use=use, **kwargs)

        try:
            radiate = lat.radiate
        except AttributeError:
            radiate = False
        else:
            del lat.radiate
        # noinspection PyUnboundLocalVariable
        if radiate:
            lat.enable_6d(copy=False)
        return lat


class MadxParser(_MadParser):
    # noinspection PyUnresolvedReferences
    """MAD-X specific parser

    The parser is a subclass of :py:class:`dict` and is database containing all the
    MAD-X parameters and objects.

    Example:
        Parse a 1st file:

        >>> parser = at.MadxParser()
        >>> parser.parse_file("file1")

        Parse another file. This adds the new definitions to the database:

        >>> parser.parse_file("file2")

        Look at the *rf_on* variable

        >>> parser["rf_on"]
        0

        Get the list of available sequences/lines:

        >>> parser.sequences
        ['arca',
         'arca_inj',
         'arcb_inj',
         'low_emit_ring',
         'arc_inj',
         'low_emit_ring_inj']

        Generate an AT :py:class:`.Lattice` from the *low_emit_ring* sequence

        >>> ring1 = parser.lattice(use="low_emit_ring")

        Change the value of a variable:

        >>> parser["rf_on"] = 1

        Generate a new AT :py:class:`.Lattice` taking into account the new variables:

        >>> ring2 = parser.lattice(use="low_emit_ring")

        Generate an AT :py:class:`.Lattice` from another sequence:

        >>> arca = parser.lattice(use="ring")
    """

    _continuation = None
    _blockcomment = ("/*", "*/")

    def __init__(self, **kwargs):
        """
        Args:
            strict:     If :py:obj:`False`, assign 0 to undefined variables
            verbose:    If :py:obj:`True`, print details on the processing
            **kwargs:   Initial variable definitions
        """
        super().__init__(
            globals(),
            **kwargs,
        )

    def _format_command(self, expr: str) -> str:
        """Format a command for evaluation"""
        expr = expr.replace("->", ".")  # Attribute access: VAR->ATTR
        expr = expr.replace("^", "**")  # Exponentiation
        return super()._format_command(expr)


def load_madx(
    *files: str,
    use: str = "ring",
    strict: bool = True,
    verbose: bool = False,
    **kwargs,
) -> Lattice:
    """Create a :py:class:`.Lattice` from MAD-X files

    - The *energy* and *particle* of the generated lattice are taken from the MAD-X
      ``BEAM`` object, using the MAD-X default parameters: positrons at 1 Gev.
      These parameters are overloaded by the value given in the *energy* and
      *particle* keyword arguments.
    - The radiation state is given by the ``RADIATE`` flag of the ``BEAM`` object,
      using the AT defaults: RF cavities active, synchrotron radiation in dipoles and
      quadrupoles.
    - Long elements are split according to the default AT value of *NumIntSteps* (10).

    Parameters:
        files:              Names of one or several MAD-X files
        strict:             If :py:obj:`False`, assign 0 to undefined variables
        use:                Name of the MADX sequence or line containing the desired
          lattice. Default: ``ring``
        verbose:            If :py:obj:`True`, print details on the processing

    Keyword Args:
        name (str):         Name of the lattice. Default: MADX sequence name.
        particle(Particle): Circulating particle. Default: from MADX
        energy (float):     Energy of the lattice [eV]. Default: from MADX
        periodicity(int):   Number of periods. Default: 1
        *:                  Other keywords will be used as initial variable definitions

    Returns:
        lattice (Lattice):  New :py:class:`.Lattice` object

    See Also:
        :py:func:`.load_lattice` for a generic lattice-loading function.
    """
    parser = MadxParser(strict=strict, verbose=verbose)
    params = {
        key: kwargs.pop(key)
        for key in ("name", "particle", "energy", "periodicity")
        if key in kwargs
    }
    parser.parse_files(*files, **kwargs)
    return parser.lattice(use=use, **params)


def longmultipole(kwargs):
    length = kwargs.get("Length", 0.0)
    if length == 0.0:
        return multipole.from_at(kwargs)
    else:
        drname = kwargs["FamName"] + ".1"
        dr1 = drift.from_at({"FamName": drname, "Length": 0.5 * length})
        dr2 = drift.from_at({"FamName": drname, "Length": 0.5 * length})
        return [dr1, multipole.from_at(kwargs, factor=length), dr2]


def ignore(kwargs):
    length = kwargs.get("Length", 0.0)
    if length == 0.0:
        print(f"{kwargs['name']} is replaced by a marker")
        return marker.from_at(kwargs)
    else:
        print(f"{kwargs['name']} is replaced by a drift")
        return drift.from_at(kwargs)


_AT2MAD = {
    elt.Quadrupole: quadrupole.from_at,
    elt.Sextupole: sextupole.from_at,
    elt.Octupole: octupole.from_at,
    elt.ThinMultipole: multipole.from_at,
    elt.Multipole: longmultipole,
    elt.RFCavity: rfcavity.from_at,
    elt.Drift: drift.from_at,
    elt.Bend: sbend.from_at,
    elt.Marker: marker.from_at,
    elt.Monitor: monitor.from_at,
    elt.Corrector: kicker.from_at,
}


class _MadExporter(Exporter):
    use_line = False

    def generate_madelems(
        self, eltype: type[elt.Element], elemdict: dict
    ) -> ElementDescr | list[ElementDescr]:
        return _AT2MAD.get(eltype, ignore)(elemdict)

    def print_beam(self, file):
        part = str(self.particle)
        if part == "relativistic":
            part = "electron"
        data = {
            "ENERGY": 1.0e-9 * self.energy,
            "PARTICLE": part.upper(),
            "RADIATE": self.is_6d,
        }
        attrs = [f"{k}={ElementDescr.attr_format(v)}" for k, v in data.items()]
        line = ", ".join(["BEAM".ljust(10)] + attrs)
        print(f"{line}{self.delimiter}", file=file)


class _MadxExporter(_MadExporter):
    delimiter = ";"
    continuation = ""
    bool_fmt = {False: "FALSE", True: "TRUE"}


def save_madx(ring: Lattice, filename: str | None = None, **kwargs):
    """Save a :py:class:`.Lattice` as a MAD-X file

    Args:
        ring:   lattice
        filename: file to be created. If None, write to sys.stdout

    Keyword Args:
        use (str | None): name of the created SEQUENCE of LINE.
          Default: name of the PyAT lattice
        use_line (bool):  If True, use a MAD "LINE" format. Otherwise, use
          a MAD "SEQUENCE"
    """
    exporter = _MadxExporter(ring, **kwargs)
    exporter.export(filename)


register_format(
    ".seq",
    load_madx,
    save_madx,
    descr="MAD-X lattice description. See :py:func:`.load_madx`.",
)



"""
Import/export AT lattice from/to different formats:
- .mat files
- .m files
"""
from .rpn import *  # noqa: F401, F403
from .allfiles import *  # noqa: F401, F403
from .matfile import *  # noqa: F401, F403
from .reprfile import *  # noqa: F401, F403
from .tracy import *  # noqa: F401, F403
from .elegant import *  # noqa: F401, F403
from .json import *  # noqa: F401, F403
from .file_input import *  # noqa: F401, F403
from .file_output import *
from .madx import *  # noqa: F401, F403
from .mad8 import *  # noqa: F401, F403



"""Stub file for the 'gpu' extension."""

from __future__ import annotations

import numpy as np

from at.lattice import Element, Particle

def gpuinfo(): ...
def gpupass(
    line: list[Element],
    r_in: np.ndarray,
    nturns: int,
    refpts: np.ndarray = ...,
    turn: int | None = None,
    energy: float | None = None,
    particle: Particle | None = None,
    keep_counter: bool = False,
    reuse: bool = False,
    losses: bool = False,
    bunch_spos=None,
    bunch_current=None,
    gpu_pool: list[int] | None = None,
    tracking_starts=None,
    integrator=4,
    verbose: bool = False
): ...
